"""Handle common datasets used in optical flow estimation."""

# =============================================================================
# Copyright 2021 Henrique Morimitsu
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# =============================================================================

import logging
from pathlib import Path
from typing import Callable, Dict, List, Optional, Sequence, Tuple, Union

import cv2
import numpy as np
import torch
from torch.utils.data import Dataset
from ptlflow.utils import flow_utils
from ptlflow.utils.utils import config_logging

config_logging()

THIS_DIR = Path(__file__).resolve().parent


class BaseFlowDataset(Dataset):
    """Manage optical flow dataset loading.

    This class can be used as the parent for any concrete dataset. It is structured to be able to read most types of inputs
    used in optical flow estimation.

    Classes inheriting from this one should implement the __init__() method and properly load the input paths from the chosen
    dataset. This should be done by populating the lists defined in the attributes below.

    Attributes
    ----------
    img_paths : list[list[str]]
        Paths of the images. Each element of the main list is a list of paths. Typically, the inner list will have two
        elements, corresponding to the paths of two consecutive images, which will be used to estimate the optical flow.
        More than two paths can also be added in case the model is able to use more images for estimating the flow.
    flow_paths : list[list[str]]
        Similar structure to img_paths. However, the inner list must have exactly one element less than img_paths.
        For example, if an entry of img_paths is composed of two paths, then an entry of flow_list should be a list with a
        single path, corresponding to the optical flow from the first image to the second.
    occ_paths : list[list[str]]
        Paths to the occlusion masks, follows the same structure as flow_paths. It can be left empty if not available.
    mb_paths : list[list[str]]
        Paths to the motion boundary masks, follows the same structure as flow_paths. It can be left empty if not available.
    flow_b_paths : list[list[str]]
        The same as flow_paths, but it corresponds to the backward flow. This list must be in the same order as flow_paths.
        For example, flow_b_paths[i] must be backward flow of flow_paths[i]. It can be left empty if backard flows are not
        available.
    occ_b_paths : list[list[str]]
        Backward occlusion mask paths, read occ_paths and flow_b_paths above.
    mb_b_paths : list[list[str]]
        Backward motion boundary mask paths, read mb_paths and flow_b_paths above.
    metadata : list[Any]
        Some metadata for each input. It can include anything. A good recommendation would be to put a dict with the metadata.
    """

    def __init__(
        self,
        dataset_name: str,
        split_name: str = '',
        transform: Callable[[Dict[str, torch.Tensor]], Dict[str, torch.Tensor]] = None,
        max_flow: float = 10000.0,
        get_valid_mask: bool = True,
        get_occlusion_mask: bool = True,
        get_motion_boundary_mask: bool = True,
        get_backward: bool = True,
        get_meta: bool = True
    ) -> None:
        """Initialize BaseFlowDataset.

        Parameters
        ----------
        dataset_name : str
            A string representing the dataset name. It is just used to be stored as metadata, so it can have any value.
        split_name : str, optional
            A string representing the split of the data. It is just used to be stored as metadata, so it can have any value.
        transform : Callable[[Dict[str, torch.Tensor]], Dict[str, torch.Tensor]], optional
            Transform to be applied on the inputs.
        max_flow : float, default 10000.0
            Maximum optical flow absolute value. Flow absolute values that go over this limit are clipped, and also marked
            as zero in the valid mask.
        get_valid_mask : bool, default True
            Whether to get or generate valid masks.
        get_occlusion_mask : bool, default True
            Whether to get occlusion masks.
        get_motion_boundary_mask : bool, default True
            Whether to get motion boundary masks.
        get_backward : bool, default True
            Whether to get the occluded version of the inputs.
        get_meta : bool, default True
            Whether to get metadata.
        """
        self.dataset_name = dataset_name
        self.split_name = split_name
        self.transform = transform
        self.max_flow = max_flow
        self.get_valid_mask = get_valid_mask
        self.get_occlusion_mask = get_occlusion_mask
        self.get_motion_boundary_mask = get_motion_boundary_mask
        self.get_backward = get_backward
        self.get_meta = get_meta

        self.img_paths = []
        self.flow_paths = []
        self.occ_paths = []
        self.mb_paths = []
        self.flow_b_paths = []
        self.occ_b_paths = []
        self.mb_b_paths = []
        self.metadata = []

    def __getitem__(  # noqa: C901
        self,
        index: int
    ) -> Dict[str, torch.Tensor]:
        """Retrieve and return one input.

        Parameters
        ----------
        index : int
            The index of the entry on the input lists.

        Returns
        -------
        Dict[str, torch.Tensor]
            The retrieved input. This dict may contain the following keys, depending on the initialization choices:
            ['images', 'flows', 'mbs', 'occs', 'valids', 'flows_b', 'mbs_b', 'occs_b', 'valids_b', 'meta'].
            Except for 'meta', all the values are 4D tensors with shape NCHW. Notice that N does not correspond to the batch
            size, but rather to the number of images of a given key. For example, typically 'images' will have N=2, and
            'flows' will have N=1, and so on. Therefore, a batch of these inputs will be a 5D tensor BNCHW.
        """
        inputs = {}

        inputs['images'] = [cv2.imread(str(path)) for path in self.img_paths[index]]

        if index < len(self.flow_paths):
            inputs['flows'], valids = self._get_flows_and_valids(self.flow_paths, index)
            if self.get_valid_mask:
                inputs['valids'] = valids

        if self.get_occlusion_mask and index < len(self.occ_paths):
            inputs['occs'] = [cv2.imread(str(path), 0)[:, :, None] for path in self.occ_paths[index]]
        if self.get_motion_boundary_mask and index < len(self.mb_paths):
            inputs['mbs'] = [cv2.imread(str(path), 0)[:, :, None] for path in self.mb_paths[index]]

        if self.get_backward:
            if index < len(self.flow_b_paths):
                inputs['flows_b'], valids_b = self._get_flows_and_valids(self.flow_b_paths, index)
                if self.get_valid_mask:
                    inputs['valids_b'] = valids_b
            if self.get_occlusion_mask and index < len(self.occ_b_paths):
                inputs['occs_b'] = [cv2.imread(str(path), 0)[:, :, None] for path in self.occ_b_paths[index]]
            if self.get_motion_boundary_mask and index < len(self.mb_b_paths):
                inputs['mbs_b'] = [cv2.imread(str(path), 0)[:, :, None] for path in self.mb_b_paths[index]]

        if self.transform is not None:
            inputs = self.transform(inputs)

        if self.get_meta:
            inputs['meta'] = {'dataset_name': self.dataset_name, 'split_name': self.split_name}
            if index < len(self.metadata):
                inputs['meta'].update(self.metadata[index])

        return inputs

    def __len__(self) -> int:
        return len(self.img_paths)

    def _get_flows_and_valids(
        self,
        flow_paths: Sequence[str],
        index: int
    ) -> Tuple[List[np.ndarray], List[Optional[np.ndarray]]]:
        flows = []
        valids = []
        for path in flow_paths[index]:
            flow = flow_utils.flow_read(path)

            nan_mask = np.isnan(flow)
            flow[nan_mask] = self.max_flow + 1

            if self.get_valid_mask:
                valid = (np.abs(flow) < self.max_flow).astype(np.uint8)*255
                valid = np.minimum(valid[:, :, 0], valid[:, :, 1])
                valids.append(valid[:, :, None])

            flow[nan_mask] = 0

            flow = np.clip(flow, -self.max_flow, self.max_flow)
            flows.append(flow)
        return flows, valids

    def _log_status(self) -> None:
        if self.__len__() == 0:
            logging.warning(
                'No samples were found for %s dataset. Be sure to update the dataset path in datasets.yml, '
                'or provide the path by the argument --[dataset_name]_root_dir.', self.dataset_name)
        else:
            logging.info('Loading %d samples from %s dataset.', self.__len__(), self.dataset_name)


class FlyingChairsDataset(BaseFlowDataset):
    """Handle the FlyingChairs dataset."""

    def __init__(
        self,
        root_dir: str,
        split: str = 'train',
        transform: Callable[[Dict[str, torch.Tensor]], Dict[str, torch.Tensor]] = None,
        max_flow: float = 10000.0,
        get_valid_mask: bool = True,
        get_meta: bool = True
    ) -> None:
        """Initialize FlyingChairsDataset.

        Parameters
        ----------
        root_dir : str
            path to the root directory of the FlyingChairs dataset.
        split : str, default 'train'
            Which split of the dataset should be loaded. It can be one of {'train', 'val', 'trainval'}.
        transform : Callable[[Dict[str, torch.Tensor]], Dict[str, torch.Tensor]], optional
            Transform to be applied on the inputs.
        max_flow : float, default 10000.0
            Maximum optical flow absolute value. Flow absolute values that go over this limit are clipped, and also marked
            as zero in the valid mask.
        get_valid_mask : bool, default True
            Whether to get or generate valid masks.
        get_meta : bool, default True
            Whether to get metadata.
        """
        super().__init__(
            dataset_name='FlyingChairs',
            split_name=split,
            transform=transform,
            max_flow=max_flow,
            get_valid_mask=get_valid_mask,
            get_occlusion_mask=False,
            get_motion_boundary_mask=False,
            get_backward=False,
            get_meta=get_meta)
        self.root_dir = root_dir
        self.split_file = THIS_DIR / 'FlyingChairs_val.txt'

        # Read data from disk
        img1_paths = sorted((Path(self.root_dir) / 'data').glob('*img1.ppm'))
        img2_paths = sorted((Path(self.root_dir) / 'data').glob('*img2.ppm'))
        flow_paths = sorted((Path(self.root_dir) / 'data').glob('*flow.flo'))

        # Sanity check
        assert len(img1_paths) == len(img2_paths), f'{len(img1_paths)} vs {len(img2_paths)}'
        assert len(img1_paths) == len(flow_paths), f'{len(img1_paths)} vs {len(flow_paths)}'

        with open(self.split_file, 'r') as f:
            val_names = f.read().strip().splitlines()

        if split == 'trainval':
            remove_names = []
        elif split == 'train':
            remove_names = val_names
        elif split == 'val':
            remove_names = [p.stem.split('_')[0] for p in img1_paths if p.stem.split('_')[0] not in val_names]

        # Keep only data from the correct split
        self.img_paths = [
            [img1_paths[i], img2_paths[i]]
            for i in range(len(img1_paths)) if img1_paths[i].stem.split('_')[0] not in remove_names]
        self.flow_paths = [
            [flow_paths[i]] for i in range(len(flow_paths)) if flow_paths[i].stem.split('_')[0] not in remove_names]
        assert len(self.img_paths) == len(self.flow_paths), f'{len(self.img_paths)} vs {len(self.flow_paths)}'

        self.metadata = [
            {'image_paths': [str(p) for p in paths], 'is_val': paths[0].stem in val_names, 'misc': ''}
            for paths in self.img_paths]

        self._log_status()


class FlyingChairs2Dataset(BaseFlowDataset):
    """Handle the FlyingChairs 2 dataset."""

    def __init__(
        self,
        root_dir: str,
        split: str = 'train',
        add_reverse: bool = True,
        transform: Callable[[Dict[str, torch.Tensor]], Dict[str, torch.Tensor]] = None,
        max_flow: float = 1000.0,
        get_valid_mask: bool = True,
        get_occlusion_mask: bool = True,
        get_motion_boundary_mask: bool = True,
        get_backward: bool = True,
        get_meta: bool = True
    ) -> None:
        """Initialize FlyingChairs2Dataset.

        Parameters
        ----------
        root_dir : str
            path to the root directory of the FlyingChairs2 dataset.
        split : str, default 'train'
            Which split of the dataset should be loaded. It can be one of {'train', 'val', 'trainval'}.
        add_reverse : bool, default True
            If True, double the number of samples by appending the backward samples as additional samples.
        transform : Callable[[Dict[str, torch.Tensor]], Dict[str, torch.Tensor]], optional
            Transform to be applied on the inputs.
        max_flow : float, default 10000.0
            Maximum optical flow absolute value. Flow absolute values that go over this limit are clipped, and also marked
            as zero in the valid mask.
        get_valid_mask : bool, default True
            Whether to get or generate valid masks.
        get_occlusion_mask : bool, default True
            Whether to get occlusion masks.
        get_motion_boundary_mask : bool, default True
            Whether to get motion boundary masks.
        get_backward : bool, default True
            Whether to get the occluded version of the inputs.
        get_meta : bool, default True
            Whether to get metadata.
        """
        super().__init__(
            dataset_name='FlyingChairs2',
            split_name=split,
            transform=transform,
            max_flow=max_flow,
            get_valid_mask=get_valid_mask,
            get_occlusion_mask=get_occlusion_mask,
            get_motion_boundary_mask=get_motion_boundary_mask,
            get_backward=get_backward,
            get_meta=get_meta)
        self.root_dir = root_dir
        self.add_reverse = add_reverse

        if split == 'train':
            dir_names = ['train']
        elif split == 'val':
            dir_names = ['val']
        else:
            dir_names = ['train', 'val']

        for dname in dir_names:
            # Read data from disk
            img1_paths = sorted((Path(self.root_dir) / dname).glob('*img_0.png'))
            img2_paths = sorted((Path(self.root_dir) / dname).glob('*img_1.png'))
            self.img_paths.extend([[img1_paths[i], img2_paths[i]] for i in range(len(img1_paths))])
            self.flow_paths.extend([[x] for x in sorted((Path(self.root_dir) / dname).glob('*flow_01.flo'))])
            self.occ_paths.extend([[x] for x in sorted((Path(self.root_dir) / dname).glob('*occ_01.png'))])
            self.mb_paths.extend([[x] for x in sorted((Path(self.root_dir) / dname).glob('*mb_01.png'))])
            if self.get_backward:
                self.flow_b_paths.extend([[x] for x in sorted((Path(self.root_dir) / dname).glob('*flow_10.flo'))])
                self.occ_b_paths.extend([[x] for x in sorted((Path(self.root_dir) / dname).glob('*occ_10.png'))])
                self.mb_b_paths.extend([[x] for x in sorted((Path(self.root_dir) / dname).glob('*mb_10.png'))])
            if self.add_reverse:
                self.img_paths.extend([[img2_paths[i], img1_paths[i]] for i in range(len(img1_paths))])
                self.flow_paths.extend([[x] for x in sorted((Path(self.root_dir) / dname).glob('*flow_10.flo'))])
                self.occ_paths.extend([[x] for x in sorted((Path(self.root_dir) / dname).glob('*occ_10.png'))])
                self.mb_paths.extend([[x] for x in sorted((Path(self.root_dir) / dname).glob('*mb_10.png'))])
                if self.get_backward:
                    self.flow_b_paths.extend([[x] for x in sorted((Path(self.root_dir) / dname).glob('*flow_01.flo'))])
                    self.occ_b_paths.extend([[x] for x in sorted((Path(self.root_dir) / dname).glob('*occ_01.png'))])
                    self.mb_b_paths.extend([[x] for x in sorted((Path(self.root_dir) / dname).glob('*mb_01.png'))])

        self.metadata = [{'image_paths': [str(p) for p in paths], 'is_val': False, 'misc': ''} for paths in self.img_paths]

        # Sanity check
        assert len(img1_paths) == len(img2_paths), f'{len(img1_paths)} vs {len(img2_paths)}'
        assert len(self.img_paths) == len(self.flow_paths), f'{len(self.img_paths)} vs {len(self.flow_paths)}'
        assert len(self.img_paths) == len(self.occ_paths), f'{len(self.img_paths)} vs {len(self.occ_paths)}'
        assert len(self.img_paths) == len(self.mb_paths), f'{len(self.img_paths)} vs {len(self.mb_paths)}'
        if self.get_backward:
            assert len(self.img_paths) == len(self.flow_b_paths), f'{len(self.img_paths)} vs {len(self.flow_b_paths)}'
            assert len(self.img_paths) == len(self.occ_b_paths), f'{len(self.img_paths)} vs {len(self.occ_b_paths)}'
            assert len(self.img_paths) == len(self.mb_b_paths), f'{len(self.img_paths)} vs {len(self.mb_b_paths)}'

        self._log_status()


class FlyingThings3DDataset(BaseFlowDataset):
    """Handle the FlyingThings3D dataset.

    Note that this only works for the complete FlyingThings3D dataset. For the subset version, use FlyingThings3DSubsetDataset.
    """

    def __init__(  # noqa: C901
        self,
        root_dir: str,
        split: str = 'train',
        pass_names: Union[str, List[str]] = 'clean',
        side_names: Union[str, List[str]] = 'left',
        add_reverse: bool = True,
        transform: Callable[[Dict[str, torch.Tensor]], Dict[str, torch.Tensor]] = None,
        max_flow: float = 1000.0,
        get_valid_mask: bool = True,
        get_occlusion_mask: bool = True,
        get_motion_boundary_mask: bool = True,
        get_backward: bool = True,
        get_meta: bool = True,
        sequence_length: int = 2
    ) -> None:
        """Initialize FlyingThings3DDataset.

        Parameters
        ----------
        root_dir : str
            path to the root directory of the FlyingThings3D dataset.
        split : str, default 'train'
            Which split of the dataset should be loaded. It can be one of {'train', 'val', 'trainval'}.
        pass_names : Union[str, List[str]], default 'clean'
            Which passes should be loaded. It can be one of {'clean', 'final', ['clean', 'final']}.
        side_names : Union[str, List[str]], default 'left'
             Samples from which side view should be loaded. It can be one of {'left', 'right', ['left', 'right']}.
        add_reverse : bool, default True
            If True, double the number of samples by appending the backward samples as additional samples.
        transform : Callable[[Dict[str, torch.Tensor]], Dict[str, torch.Tensor]], optional
            Transform to be applied on the inputs.
        max_flow : float, default 10000.0
            Maximum optical flow absolute value. Flow absolute values that go over this limit are clipped, and also marked
            as zero in the valid mask.
        get_valid_mask : bool, default True
            Whether to get or generate valid masks.
        get_occlusion_mask : bool, default True
            Whether to get occlusion masks.
        get_motion_boundary_mask : bool, default True
            Whether to get motion boundary masks.
        get_backward : bool, default True
            Whether to get the occluded version of the inputs.
        get_meta : bool, default True
            Whether to get metadata.
        sequence_length : int, default 2
            How many consecutive images are loaded per sample. More than two images can be used for model which exploit more
            temporal information.
        """
        super().__init__(
            dataset_name='FlyingThings3D',
            split_name=split,
            transform=transform,
            max_flow=max_flow,
            get_valid_mask=get_valid_mask,
            get_occlusion_mask=get_occlusion_mask,
            get_motion_boundary_mask=get_motion_boundary_mask,
            get_backward=get_backward,
            get_meta=get_meta)
        self.root_dir = root_dir
        self.add_reverse = add_reverse
        self.pass_names = pass_names
        self.sequence_length = sequence_length
        if isinstance(self.pass_names, str):
            self.pass_names = [self.pass_names]
        self.side_names = side_names
        if isinstance(self.side_names, str):
            self.side_names = [self.side_names]

        if split == 'val':
            split_dir_names = ['TEST']
        elif split == 'train':
            split_dir_names = ['TRAIN']
        else:
            split_dir_names = ['TRAIN', 'TEST']

        pass_dirs = [f'frames_{p}pass' for p in self.pass_names]

        directions = [('into_future', 'into_past')]
        reverts = [False]
        if self.add_reverse:
            directions.append(('into_past', 'into_future'))
            reverts.append(True)

        # Read paths from disk
        for passd in pass_dirs:
            for split in split_dir_names:
                split_path = Path(self.root_dir) / passd / split
                for letter_path in split_path.glob('*'):
                    for seq_path in letter_path.glob('*'):
                        for direcs, rev in zip(directions, reverts):
                            for side in self.side_names:
                                image_paths = sorted((seq_path / side).glob('*.png'), reverse=rev)
                                flow_paths = sorted(
                                    (Path(str(seq_path).replace(passd, 'optical_flow')) / direcs[0] / side).glob('*.pfm'),
                                    reverse=rev)

                                occ_paths = []
                                if (Path(self.root_dir) / 'occlusions').exists():
                                    occ_paths = sorted(
                                        (Path(str(seq_path).replace(passd, 'occlusions')) / direcs[0] / side).glob('*.png'),
                                        reverse=rev)
                                mb_paths = []
                                if (Path(self.root_dir) / 'motion_boundaries').exists():
                                    mb_paths = sorted(
                                        (Path(str(seq_path).replace(passd, 'motion_boundaries')) / direcs[0] / side).glob(
                                            '*.png'),
                                        reverse=rev)

                                flow_b_paths = []
                                occ_b_paths = []
                                mb_b_paths = []
                                if self.get_backward:
                                    flow_b_paths = sorted(
                                        (Path(str(seq_path).replace(passd, 'optical_flow')) / direcs[1] / side).glob('*.pfm'),
                                        reverse=rev)
                                    if (Path(self.root_dir) / 'occlusions').exists():
                                        occ_b_paths = sorted(
                                            (Path(str(seq_path).replace(passd, 'occlusions')) / direcs[1] / side).glob(
                                                '*.png'),
                                            reverse=rev)
                                    if (Path(self.root_dir) / 'motion_boundaries').exists():
                                        mb_b_paths = sorted(
                                            (Path(str(seq_path).replace(passd, 'motion_boundaries')) / direcs[1] / side).glob(
                                                '*.png'),
                                            reverse=rev)

                                for i in range(len(image_paths)-self.sequence_length+1):
                                    self.img_paths.append(image_paths[i:i+self.sequence_length])
                                    if len(flow_paths) > 0:
                                        self.flow_paths.append(flow_paths[i:i+self.sequence_length-1])
                                    if len(occ_paths) > 0:
                                        self.occ_paths.append(occ_paths[i:i+self.sequence_length-1])
                                    if len(mb_paths) > 0:
                                        self.mb_paths.append(mb_paths[i:i+self.sequence_length-1])
                                    self.metadata.append({
                                        'image_paths': [str(p) for p in image_paths[i:i+self.sequence_length]],
                                        'is_val': False,
                                        'misc': ''
                                    })
                                    if self.get_backward:
                                        if len(flow_b_paths) > 0:
                                            self.flow_b_paths.append(flow_b_paths[i+1:i+self.sequence_length])
                                        if len(occ_b_paths) > 0:
                                            self.occ_b_paths.append(occ_b_paths[i+1:i+self.sequence_length])
                                        if len(mb_b_paths) > 0:
                                            self.mb_b_paths.append(mb_b_paths[i+1:i+self.sequence_length])

        assert len(self.img_paths) == len(self.flow_paths), f'{len(self.img_paths)} vs {len(self.flow_paths)}'
        assert len(self.occ_paths) == 0 or len(self.img_paths) == len(self.occ_paths), (
            f'{len(self.img_paths)} vs {len(self.occ_paths)}')
        assert len(self.mb_paths) == 0 or len(self.img_paths) == len(self.mb_paths), (
            f'{len(self.img_paths)} vs {len(self.mb_paths)}')
        if self.get_backward:
            assert len(self.img_paths) == len(self.flow_b_paths), f'{len(self.img_paths)} vs {len(self.flow_b_paths)}'
            assert len(self.occ_b_paths) == 0 or len(self.img_paths) == len(self.occ_b_paths), (
                f'{len(self.img_paths)} vs {len(self.occ_b_paths)}')
            assert len(self.mb_b_paths) == 0 or len(self.img_paths) == len(self.mb_b_paths), (
                f'{len(self.img_paths)} vs {len(self.mb_b_paths)}')

        self._log_status()


class FlyingThings3DSubsetDataset(BaseFlowDataset):
    """Handle the FlyingThings3D subset dataset.

    Note that this only works for the FlyingThings3D subset dataset. For the complete version, use FlyingThings3DDataset.
    """

    def __init__(  # noqa: C901
        self,
        root_dir: str,
        split: str = 'train',
        pass_names: Union[str, List[str]] = 'clean',
        side_names: Union[str, List[str]] = 'left',
        add_reverse: bool = True,
        transform: Callable[[Dict[str, torch.Tensor]], Dict[str, torch.Tensor]] = None,
        max_flow: float = 1000.0,
        get_valid_mask: bool = True,
        get_occlusion_mask: bool = True,
        get_motion_boundary_mask: bool = True,
        get_backward: bool = True,
        get_meta: bool = True,
        sequence_length: int = 2
    ) -> None:
        """Initialize FlyingThings3DSubsetDataset.

        Parameters
        ----------
        root_dir : str
            path to the root directory of the FlyingThings3D dataset.
        split : str, default 'train'
            Which split of the dataset should be loaded. It can be one of {'train', 'val', 'trainval'}.
        pass_names : Union[str, List[str]], default 'clean'
            Which passes should be loaded. It can be one of {'clean', 'final', ['clean', 'final']}.
        side_names : Union[str, List[str]], default 'left'
             Samples from which side view should be loaded. It can be one of {'left', 'right', ['left', 'right']}.
        add_reverse : bool, default True
            If True, double the number of samples by appending the backward samples as additional samples.
        transform : Callable[[Dict[str, torch.Tensor]], Dict[str, torch.Tensor]], optional
            Transform to be applied on the inputs.
        max_flow : float, default 10000.0
            Maximum optical flow absolute value. Flow absolute values that go over this limit are clipped, and also marked
            as zero in the valid mask.
        get_valid_mask : bool, default True
            Whether to get or generate valid masks.
        get_occlusion_mask : bool, default True
            Whether to get occlusion masks.
        get_motion_boundary_mask : bool, default True
            Whether to get motion boundary masks.
        get_backward : bool, default True
            Whether to get the occluded version of the inputs.
        get_meta : bool, default True
            Whether to get metadata.
        sequence_length : int, default 2
            How many consecutive images are loaded per sample. More than two images can be used for model which exploit more
            temporal information.
        """
        super().__init__(
            dataset_name='FlyingThings3DSubset',
            split_name=split,
            transform=transform,
            max_flow=max_flow,
            get_valid_mask=get_valid_mask,
            get_occlusion_mask=get_occlusion_mask,
            get_motion_boundary_mask=get_motion_boundary_mask,
            get_backward=get_backward,
            get_meta=get_meta)
        self.root_dir = root_dir
        self.add_reverse = add_reverse
        self.pass_names = pass_names
        self.sequence_length = sequence_length
        if isinstance(self.pass_names, str):
            self.pass_names = [self.pass_names]
        self.side_names = side_names
        if isinstance(self.side_names, str):
            self.side_names = [self.side_names]

        if split == 'train' or split == 'val':
            split_dir_names = [split]
        else:
            split_dir_names = ['train', 'val']

        directions = [('into_future', 'into_past')]
        reverts = [False]
        if self.add_reverse:
            directions.append(('into_past', 'into_future'))
            reverts.append(True)

        # Read paths from disk
        for split in split_dir_names:
            for pass_name in self.pass_names:
                for side in self.side_names:
                    for direcs, rev in zip(directions, reverts):
                        flow_dir = Path(self.root_dir) / split / 'flow' / side / direcs[0]
                        flow_paths = sorted(flow_dir.glob('*.flo'), reverse=rev)

                        # Create groups to separate different sequences
                        flow_groups_paths = [[flow_paths[0]]]
                        prev_idx = int(flow_paths[0].stem)
                        for path in flow_paths[1:]:
                            idx = int(path.stem)
                            if (idx - 1) == prev_idx:
                                flow_groups_paths[-1].append(path)
                            else:
                                flow_groups_paths.append([path])
                            prev_idx = idx

                        for flow_group in flow_groups_paths:
                            for i in range(len(flow_group)-self.sequence_length+2):
                                flow_paths = flow_group[i:i+self.sequence_length-1]
                                self.flow_paths.append(flow_paths)

                                img_dir = Path(self.root_dir) / split / f'image_{pass_name}' / side
                                img_paths = [img_dir / (fp.stem+'.png') for fp in flow_paths]
                                if rev:
                                    idx = int(img_paths[0].stem) - 1
                                else:
                                    idx = int(img_paths[-1].stem) + 1
                                img_paths.append(img_dir / f'{idx:07d}.png')
                                self.img_paths.append(img_paths)

                                if (Path(self.root_dir) / split / 'flow_occlusions').exists():
                                    occ_paths = [
                                        Path(str(fp).replace('flow', 'flow_occlusions').replace('.flo', '.png'))
                                        for fp in flow_paths]
                                    self.occ_paths.append(occ_paths)
                                if (Path(self.root_dir) / split / 'motion_boundaries').exists():
                                    mb_paths = [
                                        Path(str(fp).replace('flow', 'motion_boundaries').replace('.flo', '.png'))
                                        for fp in flow_paths]
                                    self.mb_paths.append(mb_paths)

                                self.metadata.append({
                                    'image_paths': [str(p) for p in img_paths],
                                    'is_val': False,
                                    'misc': ''
                                })

                        if self.get_backward:
                            flow_dir = Path(self.root_dir) / split / 'flow' / side / direcs[1]
                            flow_paths = sorted(flow_dir.glob('*.flo'), reverse=rev)

                            # Create groups to separate different sequences
                            flow_groups_paths = [[flow_paths[0]]]
                            prev_idx = int(flow_paths[0].stem)
                            for path in flow_paths[1:]:
                                idx = int(path.stem)
                                if (idx - 1) == prev_idx:
                                    flow_groups_paths[-1].append(path)
                                else:
                                    flow_groups_paths.append([path])
                                prev_idx = idx

                            for flow_group in flow_groups_paths:
                                for i in range(len(flow_group)-self.sequence_length+2):
                                    flow_paths = flow_group[i:i+self.sequence_length-1]
                                    self.flow_b_paths.append(flow_paths)

                                    if (Path(self.root_dir) / split / 'flow_occlusions').exists():
                                        occ_paths = [
                                            Path(str(fp).replace('flow', 'flow_occlusions').replace('.flo', '.png'))
                                            for fp in flow_paths]
                                        self.occ_b_paths.append(occ_paths)
                                    if (Path(self.root_dir) / split / 'motion_boundaries').exists():
                                        mb_paths = [
                                            Path(str(fp).replace('flow', 'motion_boundaries').replace('.flo', '.png'))
                                            for fp in flow_paths]
                                        self.mb_b_paths.append(mb_paths)

        assert len(self.img_paths) == len(self.flow_paths), f'{len(self.img_paths)} vs  {len(self.flow_paths)}'
        assert len(self.occ_paths) == 0 or len(self.img_paths) == len(self.occ_paths), (
            f'{len(self.img_paths)} vs {len(self.occ_paths)}')
        assert len(self.mb_paths) == 0 or len(self.img_paths) == len(self.mb_paths), (
            f'{len(self.img_paths)} vs {len(self.mb_paths)}')
        if self.get_backward:
            assert len(self.img_paths) == len(self.flow_b_paths), f'{len(self.img_paths)} vs {len(self.flow_b_paths)}'
            assert len(self.occ_b_paths) == 0 or len(self.img_paths) == len(self.occ_b_paths), (
                f'{len(self.img_paths)} vs {len(self.occ_b_paths)}')
            assert len(self.mb_b_paths) == 0 or len(self.img_paths) == len(self.mb_b_paths), (
                f'{len(self.img_paths)} vs {len(self.mb_b_paths)}')

        self._log_status()


class Hd1kDataset(BaseFlowDataset):
    """Handle the HD1K dataset."""

    def __init__(  # noqa: C901
        self,
        root_dir: str,
        split: str = 'train',
        transform: Callable[[Dict[str, torch.Tensor]], Dict[str, torch.Tensor]] = None,
        max_flow: float = 512.0,
        get_valid_mask: bool = True,
        get_meta: bool = True,
        sequence_length: int = 2
    ) -> None:
        """Initialize Hd1kDataset.

        Parameters
        ----------
        root_dir : str
            path to the root directory of the HD1K dataset.
        split : str, default 'train'
            Which split of the dataset should be loaded. It can be one of {'train', 'val', 'trainval', 'test'}.
        transform : Callable[[Dict[str, torch.Tensor]], Dict[str, torch.Tensor]], optional
            Transform to be applied on the inputs.
        max_flow : float, default 512.0
            Maximum optical flow absolute value. Flow absolute values that go over this limit are clipped, and also marked
            as zero in the valid mask.
        get_valid_mask : bool, default True
            Whether to get or generate valid masks.
        get_meta : bool, default True
            Whether to get metadata.
        sequence_length : int, default 2
            How many consecutive images are loaded per sample. More than two images can be used for model which exploit more
            temporal information.
        """
        super().__init__(
            dataset_name='HD1K',
            split_name=split,
            transform=transform,
            max_flow=max_flow,
            get_valid_mask=get_valid_mask,
            get_occlusion_mask=False,
            get_motion_boundary_mask=False,
            get_backward=False,
            get_meta=get_meta)
        self.root_dir = root_dir
        self.split = split
        self.sequence_length = sequence_length

        if split == 'test':
            split_dir = 'hd1k_challenge'
        else:
            split_dir = 'hd1k_input'

        img_paths = sorted((Path(root_dir) / split_dir / 'image_2').glob('*.png'))
        img_names = [p.stem for p in img_paths]

        # Group paths by sequence
        img_names_grouped = {}
        for n in img_names:
            seq_name = n.split('_')[0]
            if img_names_grouped.get(seq_name) is None:
                img_names_grouped[seq_name] = []
            img_names_grouped[seq_name].append(n)

        val_names = []
        split_file = THIS_DIR / 'Hd1k_val.txt'
        with open(split_file, 'r') as f:
            val_names = f.read().strip().splitlines()

        # Remove names that do not belong to the chosen split
        for seq_name, seq_img_names in img_names_grouped.items():
            if split == 'train':
                img_names_grouped[seq_name] = [n for n in seq_img_names if n not in val_names]
            elif split == 'val':
                img_names_grouped[seq_name] = [n for n in seq_img_names if n in val_names]

        for seq_img_names in img_names_grouped.values():
            for i in range(len(seq_img_names) - self.sequence_length + 1):
                self.img_paths.append(
                    [Path(root_dir) / split_dir / 'image_2' / (n+'.png') for n in seq_img_names[i:i+self.sequence_length]])
                if split != 'test':
                    self.flow_paths.append(
                        [Path(root_dir) / 'hd1k_flow_gt' / 'flow_occ' / (n+'.png')
                         for n in seq_img_names[i:i+self.sequence_length-1]])

                self.metadata.append({
                    'image_paths': [str(p) for p in self.img_paths[-1]],
                    'is_val': (seq_img_names[i] in val_names),
                    'misc': ''
                })

        if split != 'test':
            assert len(self.img_paths) == len(self.flow_paths), f'{len(self.img_paths)} vs {len(self.flow_paths)}'

        self._log_status()


class KittiDataset(BaseFlowDataset):
    """Handle the KITTI dataset."""

    def __init__(  # noqa: C901
        self,
        root_dir_2012: Optional[str] = None,
        root_dir_2015: Optional[str] = None,
        split: str = 'train',
        versions: Union[str, List[str]] = '2015',
        transform: Callable[[Dict[str, torch.Tensor]], Dict[str, torch.Tensor]] = None,
        max_flow: float = 512.0,
        get_valid_mask: bool = True,
        get_meta: bool = True
    ) -> None:
        """Initialize KittiDataset.

        Parameters
        ----------
        root_dir_2012 : str, optional.
            Path to the root directory of the KITTI 2012 dataset, if available.
        root_dir_2015 : str, optional.
            Path to the root directory of the KITTI 2015 dataset, if available.
        split : str, default 'train'
            Which split of the dataset should be loaded. It can be one of {'train', 'val', 'trainval', 'test'}.
        versions : Union[str, List[str]], default '2015'
            Which version should be loaded. It can be one of {'2012', '2015', ['2012', '2015']}.
        transform : Callable[[Dict[str, torch.Tensor]], Dict[str, torch.Tensor]], optional
            Transform to be applied on the inputs.
        max_flow : float, default 512.0
            Maximum optical flow absolute value. Flow absolute values that go over this limit are clipped, and also marked
            as zero in the valid mask.
        get_valid_mask : bool, default True
            Whether to get or generate valid masks.
        get_meta : bool, default True
            Whether to get metadata.
        """
        if isinstance(versions, str):
            versions = [versions]
        super().__init__(
            dataset_name=f'KITTI_{"_".join(versions)}',
            split_name=split,
            transform=transform,
            max_flow=max_flow,
            get_valid_mask=get_valid_mask,
            get_occlusion_mask=False,
            get_motion_boundary_mask=False,
            get_backward=False,
            get_meta=get_meta)
        self.root_dir = {'2012': root_dir_2012, '2015': root_dir_2015}
        self.versions = versions
        self.split = split

        if split == 'test':
            split_dir = 'testing'
        else:
            split_dir = 'training'

        for ver in versions:
            if self.root_dir[ver] is None:
                continue

            if ver == '2012':
                image_dir = 'colored_0'
            else:
                image_dir = 'image_2'

            img1_paths = sorted((Path(self.root_dir[ver]) / split_dir / image_dir).glob('*_10.png'))
            img2_paths = sorted((Path(self.root_dir[ver]) / split_dir / image_dir).glob('*_11.png'))
            assert len(img1_paths) == len(img2_paths), f'{len(img1_paths)} vs {len(img2_paths)}'
            flow_paths = []
            if split != 'test':
                flow_paths = sorted((Path(self.root_dir[ver]) / split_dir / 'flow_occ').glob('*_10.png'))
                assert len(img1_paths) == len(flow_paths), f'{len(img1_paths)} vs {len(flow_paths)}'

            split_file = THIS_DIR / f'Kitti{ver}_val.txt'
            with open(split_file, 'r') as f:
                val_names = f.read().strip().splitlines()

            if split == 'trainval' or split == 'test':
                remove_names = []
            elif split == 'train':
                remove_names = val_names
            elif split == 'val':
                remove_names = [p.stem for p in img1_paths if p.stem not in val_names]

            self.img_paths = [
                [img1_paths[i], img2_paths[i]] for i in range(len(img1_paths)) if img1_paths[i].stem not in remove_names]
            if split != 'test':
                self.flow_paths = [
                    [flow_paths[i]] for i in range(len(flow_paths)) if flow_paths[i].stem not in remove_names]
            self.metadata = [
                {
                    'image_paths': [str(img1_paths[i]), str(img2_paths[i])],
                    'is_val': img1_paths[i].stem in val_names,
                    'misc': ver
                } for i in range(len(img1_paths)) if img1_paths[i].stem not in remove_names]

        if split != 'test':
            assert len(self.img_paths) == len(self.flow_paths), f'{len(self.img_paths)} vs {len(self.flow_paths)}'

        self._log_status()


class SintelDataset(BaseFlowDataset):
    """Handle the MPI Sintel dataset."""

    def __init__(  # noqa: C901
        self,
        root_dir: str,
        split: str = 'train',
        pass_names: Union[str, List[str]] = 'clean',
        transform: Callable[[Dict[str, torch.Tensor]], Dict[str, torch.Tensor]] = None,
        max_flow: float = 10000.0,
        get_valid_mask: bool = True,
        get_occlusion_mask: bool = True,
        get_meta: bool = True,
        sequence_length: int = 2
    ) -> None:
        """Initialize SintelDataset.

        Parameters
        ----------
        root_dir : str
            path to the root directory of the MPI Sintel dataset.
        split : str, default 'train'
            Which split of the dataset should be loaded. It can be one of {'train', 'val', 'trainval', 'test'}.
        pass_names : Union[str, List[str]], default 'clean'
            Which passes should be loaded. It can be one of {'clean', 'final', ['clean', 'final']}.
        transform : Callable[[Dict[str, torch.Tensor]], Dict[str, torch.Tensor]], optional
            Transform to be applied on the inputs.
        max_flow : float, default 10000.0
            Maximum optical flow absolute value. Flow absolute values that go over this limit are clipped, and also marked
            as zero in the valid mask.
        get_valid_mask : bool, default True
            Whether to get or generate valid masks.
        get_occlusion_mask : bool, default True
            Whether to get occlusion masks.
        get_meta : bool, default True
            Whether to get metadata.
        sequence_length : int, default 2
            How many consecutive images are loaded per sample. More than two images can be used for model which exploit more
            temporal information.
        """
        if isinstance(pass_names, str):
            pass_names = [pass_names]
        super().__init__(
            dataset_name=f'Sintel_{"_".join(pass_names)}',
            split_name=split,
            transform=transform,
            max_flow=max_flow,
            get_valid_mask=get_valid_mask,
            get_occlusion_mask=get_occlusion_mask,
            get_motion_boundary_mask=False,
            get_backward=False,
            get_meta=get_meta)
        self.root_dir = root_dir
        self.split = split
        self.pass_names = pass_names
        self.sequence_length = sequence_length

        # Get sequence names for the given split
        if split == 'test':
            split_dir = 'test'
        else:
            split_dir = 'training'

        split_file = THIS_DIR / 'Sintel_val.txt'
        with open(split_file, 'r') as f:
            val_seqs = f.read().strip().splitlines()

        sequence_names = sorted([p.stem for p in (Path(root_dir) / split_dir / 'clean').glob('*')])
        if split == 'train' or split == 'val':
            if split == 'train':
                sequence_names = [s for s in sequence_names if s not in val_seqs]
            else:
                sequence_names = val_seqs

        # Read paths from disk
        for passd in pass_names:
            for seq_name in sequence_names:
                image_paths = sorted((Path(self.root_dir) / split_dir / passd / seq_name).glob('*.png'))
                flow_paths = []
                occ_paths = []
                if split != 'test':
                    flow_paths = sorted((Path(self.root_dir) / split_dir / 'flow' / seq_name).glob('*.flo'))
                    assert len(image_paths)-1 == len(flow_paths), (
                        f'{passd}, {seq_name}: {len(image_paths)-1} vs {len(flow_paths)}')
                    if (Path(self.root_dir) / split_dir / 'occlusions').exists():
                        occ_paths = sorted((Path(self.root_dir) / split_dir / 'occlusions' / seq_name).glob('*.png'))
                        assert len(occ_paths) == len(flow_paths)
                for i in range(len(image_paths)-self.sequence_length+1):
                    self.img_paths.append(image_paths[i:i+self.sequence_length])
                    if len(flow_paths) > 0:
                        self.flow_paths.append(flow_paths[i:i+self.sequence_length-1])
                    if len(occ_paths) > 0:
                        self.occ_paths.append(occ_paths[i:i+self.sequence_length-1])
                    self.metadata.append({
                        'image_paths': [str(p) for p in image_paths[i:i+self.sequence_length]],
                        'is_val': seq_name in val_seqs,
                        'misc': seq_name
                    })

        # Sanity check
        if split != 'test':
            assert len(self.img_paths) == len(self.flow_paths), f'{len(self.img_paths)} vs {len(self.flow_paths)}'
        if len(self.occ_paths) > 0:
            assert len(self.img_paths) == len(self.occ_paths), f'{len(self.img_paths)} vs {len(self.occ_paths)}'

        self._log_status()



from argparse import ArgumentParser, Namespace
from pathlib import Path

import torch
import torch.nn as nn

from .pwc_modules import conv, upsample2d_as, rescale_flow, initialize_msra, compute_cost_volume
from .pwc_modules import WarpingLayer, FeatureExtractor, ContextNetwork, FlowEstimatorDense, OccContextNetwork, OccEstimatorDense
from .irr_modules import OccUpsampleNetwork, RefineFlow, RefineOcc
from ..base_model.base_model import BaseModel
from .losses import MultiScaleEPE_PWC_Bi_Occ_upsample


class IRRPWC(BaseModel):
    pretrained_checkpoints = {
        'chairs_occ': 'https://github.com/hmorimitsu/ptlflow/releases/download/weights1/irr_pwc-chairs_occ-02066cc4.ckpt',
        'things': 'https://github.com/hmorimitsu/ptlflow/releases/download/weights1/irr_pwc-things-c143e848.ckpt',
        'sintel': 'https://github.com/hmorimitsu/ptlflow/releases/download/weights1/irr_pwc-sintel-6ad65777.ckpt',
        'kitti': 'https://github.com/hmorimitsu/ptlflow/releases/download/weights1/irr_pwc-kitti-74d8868f.ckpt'
    }

    def __init__(self, args):
        super(IRRPWC, self).__init__(
            args=args,
            loss_fn=MultiScaleEPE_PWC_Bi_Occ_upsample(args),
            output_stride=64)
        self.leakyRELU = nn.LeakyReLU(0.1, inplace=True)

        self.feature_pyramid_extractor = FeatureExtractor(self.args.num_chs)
        self.warping_layer = WarpingLayer()

        self.dim_corr = (self.args.search_range * 2 + 1) ** 2
        self.num_ch_in_flo = self.dim_corr + 32 + 2
        self.num_ch_in_occ = self.dim_corr + 32 + 1

        self.flow_estimators = FlowEstimatorDense(self.num_ch_in_flo)
        self.context_networks = ContextNetwork(self.num_ch_in_flo + 448 + 2)
        self.occ_estimators = OccEstimatorDense(self.num_ch_in_occ)
        self.occ_context_networks = OccContextNetwork(self.num_ch_in_occ + 448 + 1)
        self.occ_shuffle_upsample = OccUpsampleNetwork(11, 1)

        self.conv_1x1 = nn.ModuleList([conv(196, 32, kernel_size=1, stride=1, dilation=1),
                                       conv(128, 32, kernel_size=1, stride=1, dilation=1),
                                       conv(96, 32, kernel_size=1, stride=1, dilation=1),
                                       conv(64, 32, kernel_size=1, stride=1, dilation=1)])

        self.conv_1x1_1 = conv(16, 3, kernel_size=1, stride=1, dilation=1)

        self.refine_flow = RefineFlow(2 + 1 + 32)
        self.refine_occ = RefineOcc(1 + 32 + 32)

        self.corr_params = {"pad_size": self.args.search_range, "kernel_size": 1, "max_disp": self.args.search_range, "stride1": 1, "stride2": 1, "corr_multiply": 1}

        initialize_msra(self.modules())

    @staticmethod
    def add_model_specific_args(parent_parser=None):
        parent_parser = BaseModel.add_model_specific_args(parent_parser)
        parser = ArgumentParser(parents=[parent_parser], add_help=False)
        parser.add_argument('--div_flow', type=float, default=0.05)
        parser.add_argument('--search_range', type=int, default=4)
        parser.add_argument('--output_level', type=int, default=4)
        parser.add_argument('--num_levels', type=int, default=7)
        parser.add_argument('--num_chs', type=int, nargs='+', default=[3, 16, 32, 64, 96, 128, 196])
        parser.add_argument('--num_iters', type=int, default=1)
        return parser

    def forward(self, inputs):

        x1_raw = inputs['images'][:, 0]
        x2_raw = inputs['images'][:, 1]
        batch_size, _, height_im, width_im = x1_raw.size()

        # on the bottom level are original images
        x1_pyramid = self.feature_pyramid_extractor(x1_raw) + [x1_raw]
        x2_pyramid = self.feature_pyramid_extractor(x2_raw) + [x2_raw]

        # outputs
        output_dict = {}
        output_dict_eval = {}
        flows = []
        occs = []

        _, _, h_x1, w_x1, = x1_pyramid[0].size()
        flow_f = torch.zeros(batch_size, 2, h_x1, w_x1).to(dtype=x1_raw.dtype, device=x1_raw.device)
        flow_b = torch.zeros(batch_size, 2, h_x1, w_x1).to(dtype=x1_raw.dtype, device=x1_raw.device)
        occ_f = torch.zeros(batch_size, 1, h_x1, w_x1).to(dtype=x1_raw.dtype, device=x1_raw.device)
        occ_b = torch.zeros(batch_size, 1, h_x1, w_x1).to(dtype=x1_raw.dtype, device=x1_raw.device)

        for l, (x1, x2) in enumerate(zip(x1_pyramid, x2_pyramid)):

            if l <= self.args.output_level:

                # warping
                if l == 0:
                    x2_warp = x2
                    x1_warp = x1
                else:
                    flow_f = upsample2d_as(flow_f, x1, mode="bilinear")
                    flow_b = upsample2d_as(flow_b, x2, mode="bilinear")
                    occ_f = upsample2d_as(occ_f, x1, mode="bilinear")
                    occ_b = upsample2d_as(occ_b, x2, mode="bilinear")
                    x2_warp = self.warping_layer(x2, flow_f, height_im, width_im, self.args.div_flow)
                    x1_warp = self.warping_layer(x1, flow_b, height_im, width_im, self.args.div_flow)

                # correlation
                out_corr_f = compute_cost_volume(x1, x2_warp, self.corr_params)
                out_corr_b = compute_cost_volume(x2, x1_warp, self.corr_params)

                out_corr_relu_f = self.leakyRELU(out_corr_f)
                out_corr_relu_b = self.leakyRELU(out_corr_b)

                if l != self.args.output_level:
                    x1_1by1 = self.conv_1x1[l](x1)
                    x2_1by1 = self.conv_1x1[l](x2)
                else:
                    x1_1by1 = x1
                    x2_1by1 = x2

                # concat and estimate flow
                flow_f = rescale_flow(flow_f, self.args.div_flow, width_im, height_im, to_local=True)
                flow_b = rescale_flow(flow_b, self.args.div_flow, width_im, height_im, to_local=True)

                x_intm_f, flow_res_f = self.flow_estimators(torch.cat([out_corr_relu_f, x1_1by1, flow_f], dim=1))
                x_intm_b, flow_res_b = self.flow_estimators(torch.cat([out_corr_relu_b, x2_1by1, flow_b], dim=1))
                flow_est_f = flow_f + flow_res_f
                flow_est_b = flow_b + flow_res_b

                flow_cont_f = flow_est_f + self.context_networks(torch.cat([x_intm_f, flow_est_f], dim=1))
                flow_cont_b = flow_est_b + self.context_networks(torch.cat([x_intm_b, flow_est_b], dim=1))

                # occ estimation
                x_intm_occ_f, occ_res_f = self.occ_estimators(torch.cat([out_corr_relu_f, x1_1by1, occ_f], dim=1))
                x_intm_occ_b, occ_res_b = self.occ_estimators(torch.cat([out_corr_relu_b, x2_1by1, occ_b], dim=1))
                occ_est_f = occ_f + occ_res_f
                occ_est_b = occ_b + occ_res_b

                occ_cont_f = occ_est_f + self.occ_context_networks(torch.cat([x_intm_occ_f, occ_est_f], dim=1))
                occ_cont_b = occ_est_b + self.occ_context_networks(torch.cat([x_intm_occ_b, occ_est_b], dim=1))

                # refinement
                img1_resize = upsample2d_as(x1_raw, flow_f, mode="bilinear")
                img2_resize = upsample2d_as(x2_raw, flow_b, mode="bilinear")
                flow_cont_f = rescale_flow(flow_cont_f, self.args.div_flow, width_im, height_im, to_local=False)
                flow_cont_b = rescale_flow(flow_cont_b, self.args.div_flow, width_im, height_im, to_local=False)
                img2_warp = self.warping_layer(img2_resize, flow_cont_f, height_im, width_im, self.args.div_flow)
                img1_warp = self.warping_layer(img1_resize, flow_cont_b, height_im, width_im, self.args.div_flow)
                
                # flow refine
                flow_f = self.refine_flow(flow_cont_f.detach(), img1_resize - img2_warp, x1_1by1)
                flow_b = self.refine_flow(flow_cont_b.detach(), img2_resize - img1_warp, x2_1by1)

                flow_f = rescale_flow(flow_f, self.args.div_flow, width_im, height_im, to_local=False)
                flow_b = rescale_flow(flow_b, self.args.div_flow, width_im, height_im, to_local=False)

                # occ refine
                x2_1by1_warp = self.warping_layer(x2_1by1, flow_f, height_im, width_im, self.args.div_flow)
                x1_1by1_warp = self.warping_layer(x1_1by1, flow_b, height_im, width_im, self.args.div_flow)

                occ_f = self.refine_occ(occ_cont_f.detach(), x1_1by1, x1_1by1 - x2_1by1_warp)
                occ_b = self.refine_occ(occ_cont_b.detach(), x2_1by1, x2_1by1 - x1_1by1_warp)

                flows.append([flow_cont_f, flow_cont_b, flow_f, flow_b])
                occs.append([occ_cont_f, occ_cont_b, occ_f, occ_b])

            else:
                flow_f = upsample2d_as(flow_f, x1, mode="bilinear")
                flow_b = upsample2d_as(flow_b, x2, mode="bilinear")
                flows.append([flow_f, flow_b])

                x2_warp = self.warping_layer(x2, flow_f, height_im, width_im, self.args.div_flow)
                x1_warp = self.warping_layer(x1, flow_b, height_im, width_im, self.args.div_flow)
                flow_b_warp = self.warping_layer(flow_b, flow_f, height_im, width_im, self.args.div_flow)
                flow_f_warp = self.warping_layer(flow_f, flow_b, height_im, width_im, self.args.div_flow)

                if l != self.args.num_levels-1:
                    x1_in = self.conv_1x1_1(x1)
                    x2_in = self.conv_1x1_1(x2)
                    x1_w_in = self.conv_1x1_1(x1_warp)
                    x2_w_in = self.conv_1x1_1(x2_warp)
                else:
                    x1_in = x1
                    x2_in = x2
                    x1_w_in = x1_warp
                    x2_w_in = x2_warp

                occ_f = self.occ_shuffle_upsample(occ_f, torch.cat([x1_in, x2_w_in, flow_f, flow_b_warp], dim=1))
                occ_b = self.occ_shuffle_upsample(occ_b, torch.cat([x2_in, x1_w_in, flow_b, flow_f_warp], dim=1))

                occs.append([occ_f, occ_b])

        outputs = {}
        if self.training:
            outputs['flow_preds'] = flows
            outputs['occ_preds'] = occs
            outputs['flows'] = (upsample2d_as(flow_f, x1_raw, mode="bilinear") * (1.0 / self.args.div_flow))[:, None]
            outputs['occs'] = upsample2d_as(torch.sigmoid(occ_f), x1_raw, mode="bilinear")[:, None]
            outputs['flows_b'] = (upsample2d_as(flow_b, x1_raw, mode="bilinear") * (1.0 / self.args.div_flow))[:, None]
            outputs['occs_b'] = upsample2d_as(torch.sigmoid(occ_b), x1_raw, mode="bilinear")[:, None]
        else:
            outputs['flows'] = (upsample2d_as(flow_f, x1_raw, mode="bilinear") * (1.0 / self.args.div_flow))[:, None]
            outputs['occs'] = upsample2d_as(torch.sigmoid(occ_f), x1_raw, mode="bilinear")[:, None]
            outputs['flows_b'] = (upsample2d_as(flow_b, x1_raw, mode="bilinear") * (1.0 / self.args.div_flow))[:, None]
            outputs['occs_b'] = upsample2d_as(torch.sigmoid(occ_b), x1_raw, mode="bilinear")[:, None]
        return outputs



from argparse import Namespace
from pathlib import Path

import torch
import torch.nn as nn
from torch.nn import init

from .flownetc import FlowNetC
from .flownets import FlowNetS
from .submodules import *
from .flownet_base import FlowNetBase

class FlowNetCSS(FlowNetBase):
    pretrained_checkpoints = {
        'things': 'https://github.com/hmorimitsu/ptlflow/releases/download/weights1/flownetcss-things-b42e67d0.ckpt'
    }

    def __init__(self, args: Namespace):
        args.input_channels = 12
        super(FlowNetCSS,self).__init__(args)

        self.args = args
        self.rgb_max = 1

        # First Block (FlowNetC)
        self.flownetc = FlowNetC(args)

        # Block (FlowNetS)
        self.flownets_1 = FlowNetS(args)
        self.flownets_2 = FlowNetS(args)

        for m in self.modules():
            if isinstance(m, nn.Conv2d):
                if m.bias is not None:
                    init.uniform_(m.bias)
                init.xavier_uniform_(m.weight)

            if isinstance(m, nn.ConvTranspose2d):
                if m.bias is not None:
                    init.uniform_(m.bias)
                init.xavier_uniform_(m.weight)
                # init_deconv_bilinear(m.weight)

    def warp(self, x, flo):
        """
        warp an image/tensor (im2) back to im1, according to the optical flow
        x: [B, C, H, W] (im2)
        flo: [B, 2, H, W] flow
        """
        B, C, H, W = x.size()
        # mesh grid 
        xx = torch.arange(0, W).view(1,-1).repeat(H,1)
        yy = torch.arange(0, H).view(-1,1).repeat(1,W)
        xx = xx.view(1,1,H,W).repeat(B,1,1,1)
        yy = yy.view(1,1,H,W).repeat(B,1,1,1)
        grid = torch.cat((xx,yy),1).float()

        if x.is_cuda:
            grid = grid.to(dtype=x.dtype, device=x.device)
        vgrid = grid + flo

        # scale grid to [-1,1] 
        vgrid[:,0,:,:] = 2.0*vgrid[:,0,:,:].clone() / max(W-1,1)-1.0
        vgrid[:,1,:,:] = 2.0*vgrid[:,1,:,:].clone() / max(H-1,1)-1.0

        vgrid = vgrid.permute(0,2,3,1)        
        output = nn.functional.grid_sample(x, vgrid, align_corners=True)
        mask = torch.ones(x.size()).to(dtype=x.dtype, device=x.device)
        mask = nn.functional.grid_sample(mask, vgrid, align_corners=True)
        
        mask[mask<0.9999] = 0
        mask[mask>0] = 1
        
        return output*mask

    def normalize(self, im):
        im = im - 0.5
        im = im / 0.5
        return im

    def forward(self, inputs):
        # flownetc
        flownetc_flow = self.flownetc(inputs)['flows'][:, 0]
        
        # warp img1 to img0; magnitude of diff between img0 and and warped_img1, 
        resampled_img1 = self.warp(inputs['images'][:, 1], flownetc_flow)
        diff_img0 = inputs['images'][:, 0] - resampled_img1 
        norm_diff_img0 = torch.norm(diff_img0, p=2, dim=1, keepdim=True)

        # concat img0, img1, img1->img0, flow, diff-mag ; 
        concat1 = torch.cat((inputs['images'][:, 0], inputs['images'][:, 1], resampled_img1, flownetc_flow/self.args.div_flow, norm_diff_img0), dim=1)
        
        # flownets1
        flownets1_flow = self.flownets_1({'images': concat1[:, None]})['flows'][:, 0]

        # warp img1 to img0; magnitude of diff between img0 and and warped_img1, 
        resampled_img1 = self.warp(inputs['images'][:, 1], flownets1_flow)
        diff_img0 = inputs['images'][:, 0] - resampled_img1 
        norm_diff_img0 = torch.norm(diff_img0, p=2, dim=1, keepdim=True)

        # concat img0, img1, img1->img0, flow, diff-mag ; 
        concat1 = torch.cat((inputs['images'][:, 0], inputs['images'][:, 1], resampled_img1, flownets1_flow/self.args.div_flow, norm_diff_img0), dim=1)
        
        # flownets2
        flownets2_preds = self.flownets_2({'images': concat1[:, None]})

        return flownets2_preds



from argparse import ArgumentParser, Namespace
from pathlib import Path

import pytorch_lightning as pl
import torch
import torch.nn as nn
import torch.nn.functional as F

from .update import BasicUpdateBlock, SmallUpdateBlock
from .extractor import BasicEncoder, SmallEncoder
from .corr import CorrBlock
from ..base_model.base_model import BaseModel


class SequenceLoss(nn.Module):
    def __init__(self, args):
        super().__init__()
        self.gamma = args.gamma
        self.max_flow = args.max_flow

    def forward(self, outputs, inputs):
        """ Loss function defined over sequence of flow predictions """

        flow_preds = outputs['flow_preds']
        flow_gt = inputs['flows'][:, 0]
        valid = inputs['valids'][:, 0]

        n_predictions = len(flow_preds)    
        flow_loss = 0.0

        # exlude invalid pixels and extremely large diplacements
        mag = torch.sum(flow_gt**2, dim=1, keepdim=True).sqrt()
        valid = (valid >= 0.5) & (mag < self.max_flow)

        for i in range(n_predictions):
            i_weight = self.gamma**(n_predictions - i - 1)
            i_loss = (flow_preds[i] - flow_gt).abs()
            flow_loss += i_weight * (valid * i_loss).mean()

        return flow_loss


class RAFT(BaseModel):
    pretrained_checkpoints = {
        'chairs': 'https://github.com/hmorimitsu/ptlflow/releases/download/weights1/raft-chairs-590f38f7.ckpt',
        'things': 'https://github.com/hmorimitsu/ptlflow/releases/download/weights1/raft-things-802bbcfd.ckpt',
        'sintel': 'https://github.com/hmorimitsu/ptlflow/releases/download/weights1/raft-sintel-fb44381e.ckpt',
        'kitti': 'https://github.com/hmorimitsu/ptlflow/releases/download/weights1/raft-kitti-3a831a4b.ckpt'
    }

    def __init__(self,
                 args: Namespace) -> None:
        super().__init__(
            args=args,
            loss_fn=SequenceLoss(args),
            output_stride=8)

        self.hidden_dim = hdim = 128
        self.context_dim = cdim = 128
        args.corr_levels = 4
        args.corr_radius = 4

        if 'dropout' not in self.args:
            self.args.dropout = 0

        # feature network, context network, and update block
        self.fnet = BasicEncoder(output_dim=256, norm_fn='instance', dropout=args.dropout)        
        self.cnet = BasicEncoder(output_dim=hdim+cdim, norm_fn='batch', dropout=args.dropout)
        self.update_block = BasicUpdateBlock(args, hidden_dim=hdim)

    @staticmethod
    def add_model_specific_args(parent_parser=None):
        parent_parser = BaseModel.add_model_specific_args(parent_parser)
        parser = ArgumentParser(parents=[parent_parser], add_help=False)
        parser.add_argument('--corr_levels', type=int, default=4)
        parser.add_argument('--corr_radius', type=int, default=4)
        parser.add_argument('--dropout', type=float, default=0.0)
        parser.add_argument('--gamma', type=float, default=0.8)
        parser.add_argument('--max_flow', type=float, default=1000.0)
        parser.add_argument('--iters', type=int, default=12)
        return parser

    def freeze_bn(self):
        for m in self.modules():
            if isinstance(m, nn.BatchNorm2d):
                m.eval()

    def coords_grid(self, batch, ht, wd):
        coords = torch.meshgrid(torch.arange(ht, dtype=self.dtype, device=self.device), torch.arange(wd, dtype=self.dtype, device=self.device))
        coords = torch.stack(coords[::-1], dim=0).float()
        return coords[None].repeat(batch, 1, 1, 1)

    def initialize_flow(self, img):
        """ Flow is represented as difference between two coordinate grids flow = coords1 - coords0"""
        N, C, H, W = img.shape
        coords0 = self.coords_grid(N, H//8, W//8)
        coords1 = self.coords_grid(N, H//8, W//8)

        # optical flow computed as difference: flow = coords1 - coords0
        return coords0, coords1

    def upflow8(self, flow, mode='bilinear'):
        new_size = (8 * flow.shape[2], 8 * flow.shape[3])
        return  8 * F.interpolate(flow, size=new_size, mode=mode, align_corners=True)

    def upsample_flow(self, flow, mask):
        """ Upsample flow field [H/8, W/8, 2] -> [H, W, 2] using convex combination """
        N, _, H, W = flow.shape
        mask = mask.view(N, 1, 9, 8, 8, H, W)
        mask = torch.softmax(mask, dim=2)

        up_flow = F.unfold(8 * flow, [3,3], padding=1)
        up_flow = up_flow.view(N, 2, 9, 1, 1, H, W)

        up_flow = torch.sum(mask * up_flow, dim=2)
        up_flow = up_flow.permute(0, 1, 4, 2, 5, 3)
        return up_flow.reshape(N, 2, 8*H, 8*W)


    def forward(self, inputs, flow_init=None):
        """ Estimate optical flow between pair of frames """
        image1 = inputs['images'][:, 0]
        image2 = inputs['images'][:, 1]

        image1 = 2 * image1 - 1.0
        image2 = 2 * image2 - 1.0

        image1 = image1.contiguous()
        image2 = image2.contiguous()

        hdim = self.hidden_dim
        cdim = self.context_dim

        # run the feature network
        fmap1, fmap2 = self.fnet([image1, image2])        
        
        fmap1 = fmap1.float()
        fmap2 = fmap2.float()
        corr_fn = CorrBlock(fmap1, fmap2, num_levels=self.args.corr_levels, radius=self.args.corr_radius)

        # run the context network
        cnet = self.cnet(image1)
        net, inp = torch.split(cnet, [hdim, cdim], dim=1)
        net = torch.tanh(net)
        inp = torch.relu(inp)

        coords0, coords1 = self.initialize_flow(image1)

        if flow_init is not None:
            coords1 = coords1 + flow_init

        flow_predictions = []
        for itr in range(self.args.iters):
            coords1 = coords1.detach()
            corr = corr_fn(coords1) # index correlation volume

            flow = coords1 - coords0
            net, up_mask, delta_flow = self.update_block(net, inp, corr, flow)

            # F(t+1) = F(t) + \Delta(t)
            coords1 = coords1 + delta_flow

            # upsample predictions
            if up_mask is None:
                flow_up = self.upflow8(coords1 - coords0)
            else:
                flow_up = self.upsample_flow(coords1 - coords0, up_mask)
            
            flow_predictions.append(flow_up)

        if self.training:
            outputs = {
                'flows': flow_up[:, None],
                'flow_preds': flow_predictions
            }
        else:
            outputs = {
                'flows': flow_up[:, None],
                'flow_small': coords1 - coords0
            }
            
        return outputs


class RAFTSmall(RAFT):
    pretrained_checkpoints = {
        'things': 'https://github.com/hmorimitsu/ptlflow/releases/download/weights1/raft_small-things-b7d9f997.ckpt'
    }

    def __init__(self,
                 args: Namespace) -> None:
        super().__init__(
            args=args)

        self.hidden_dim = hdim = 96
        self.context_dim = cdim = 64
        args.corr_levels = 4
        args.corr_radius = 3

        if 'dropout' not in self.args:
            self.args.dropout = 0

        # feature network, context network, and update block
        self.fnet = SmallEncoder(output_dim=128, norm_fn='instance', dropout=args.dropout)        
        self.cnet = SmallEncoder(output_dim=hdim+cdim, norm_fn='none', dropout=args.dropout)
        self.update_block = SmallUpdateBlock(args, hidden_dim=hdim)



from argparse import ArgumentParser, Namespace
from pathlib import Path

try:
    from spatial_correlation_sampler import SpatialCorrelationSampler
except ModuleNotFoundError:
    from ptlflow.utils.correlation import IterSpatialCorrelationSampler as SpatialCorrelationSampler
import torch
import torch.nn as nn

from .pwc_modules import conv, rescale_flow, upsample2d_as, initialize_msra
from .pwc_modules import WarpingLayer, FeatureExtractor, ContextNetwork, FlowEstimatorDense
from ..base_model.base_model import BaseModel
from .losses import MultiScaleEPE_PWC

class IRRPWCNetIRR(BaseModel):
    pretrained_checkpoints = {
        'things': 'https://github.com/hmorimitsu/ptlflow/releases/download/weights1/irr_pwcnet_irr-things-41a07190.ckpt'
    }

    def __init__(self, args):
        super(IRRPWCNetIRR, self).__init__(
            args=args,
            loss_fn=MultiScaleEPE_PWC(args),
            output_stride=64)
        self.leakyRELU = nn.LeakyReLU(0.1, inplace=True)

        self.feature_pyramid_extractor = FeatureExtractor(self.args.num_chs)
        self.warping_layer = WarpingLayer()

        self.dim_corr = (self.args.search_range * 2 + 1) ** 2
        self.num_ch_in = self.dim_corr + 32 + 2

        self.flow_estimators = FlowEstimatorDense(self.num_ch_in)

        self.context_networks = ContextNetwork(self.num_ch_in + 448 + 2)

        self.conv_1x1 = nn.ModuleList([conv(196, 32, kernel_size=1, stride=1, dilation=1),
                                       conv(128, 32, kernel_size=1, stride=1, dilation=1),
                                       conv(96, 32, kernel_size=1, stride=1, dilation=1),
                                       conv(64, 32, kernel_size=1, stride=1, dilation=1),
                                       conv(32, 32, kernel_size=1, stride=1, dilation=1)])

        self.corr = SpatialCorrelationSampler(kernel_size=1, patch_size=2*self.args.search_range+1, padding=0)

        initialize_msra(self.modules())

    @staticmethod
    def add_model_specific_args(parent_parser=None):
        parent_parser = BaseModel.add_model_specific_args(parent_parser)
        parser = ArgumentParser(parents=[parent_parser], add_help=False)
        parser.add_argument('--div_flow', type=float, default=0.05)
        parser.add_argument('--search_range', type=int, default=4)
        parser.add_argument('--output_level', type=int, default=4)
        parser.add_argument('--num_chs', type=int, nargs='+', default=[3, 16, 32, 64, 96, 128, 196])
        parser.add_argument('--num_iters', type=int, default=1)
        return parser

    def forward(self, inputs):

        x1_raw = inputs['images'][:, 0]
        x2_raw = inputs['images'][:, 1]
        _, _, height_im, width_im = x1_raw.size()

        # on the bottom level are original images
        x1_pyramid = self.feature_pyramid_extractor(x1_raw) + [x1_raw]
        x2_pyramid = self.feature_pyramid_extractor(x2_raw) + [x2_raw]

        # outputs
        output_dict = {}
        flows = []

        # init
        b_size, _, h_x1, w_x1, = x1_pyramid[0].size()
        init_dtype = x1_pyramid[0].dtype
        init_device = x1_pyramid[0].device
        flow = torch.zeros(b_size, 2, h_x1, w_x1, dtype=init_dtype, device=init_device).float()

        for l, (x1, x2) in enumerate(zip(x1_pyramid, x2_pyramid)):

            # warping
            if l == 0:
                x2_warp = x2
            else:
                flow = upsample2d_as(flow, x1, mode="bilinear")
                x2_warp = self.warping_layer(x2, flow, height_im, width_im, self.args.div_flow)

            # correlation
            out_corr = self.corr(x1, x2_warp)
            out_corr = out_corr.view(out_corr.shape[0], -1, out_corr.shape[3], out_corr.shape[4])
            out_corr = out_corr / x1.shape[1]
            out_corr_relu = self.leakyRELU(out_corr)

            # concat and estimate flow
            flow = rescale_flow(flow, self.args.div_flow, width_im, height_im, to_local=True)

            x1_1by1 = self.conv_1x1[l](x1)
            x_intm, flow_res = self.flow_estimators(torch.cat([out_corr_relu, x1_1by1, flow], dim=1))
            flow = flow + flow_res

            flow_fine = self.context_networks(torch.cat([x_intm, flow], dim=1))
            flow = flow + flow_fine

            flow = rescale_flow(flow, self.args.div_flow, width_im, height_im, to_local=False)
            flows.append(flow)

            # upsampling or post-processing
            if l == self.args.output_level:
                break

        outputs = {}
        if self.training:
            outputs['flow_preds'] = flows
            outputs['flows'] = (upsample2d_as(flow, x1_raw, mode="bilinear") * (1.0 / self.args.div_flow))[:, None]
        else:
            outputs['flows'] = (upsample2d_as(flow, x1_raw, mode="bilinear") * (1.0 / self.args.div_flow))[:, None]
        return outputs



"""Handle and compute metrics for optical flow and related estimations.

This handler is designed according to the torchmetrics specifications. Besides accuracy metrics for optical flow, it can also
compute basic metrics for occlusion, motion boundary and flow confidence estimations.
"""

# =============================================================================
# Copyright 2021 Henrique Morimitsu
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# =============================================================================

from typing import Dict

from torchmetrics import Metric
import torch


class FlowMetrics(Metric):
    """Handler for optical flow and related metrics.

    Attributes
    ----------
    average_mode : str, default 'epoch_mean'
        How the final metric is averaged. It can be either 'epoch_mean' or 'ema' (exponential moving average).
    ema_decay : float, default 0.99
        The decay to be applied if average_mode is 'ema'.
    prefix : str, optional
        A prefix string that will be attached to the metric names.
    """

    def __init__(
        self,
        dist_sync_on_step: bool = False,
        prefix: str = '',
        average_mode: str = 'epoch_mean',
        ema_decay: float = 0.99,
        f1_mode: str = 'macro'
    ) -> None:
        """Initialize FlowMetrics.

        Parameters
        ----------
        dist_sync_on_step : bool, default False
            Used by torchmetrics to sync metrics between multiple processes.
        prefix : str, optional
            A prefix string that will be attached to the metric names.
        average_mode : str, default 'epoch_mean'
            How the final metric is averaged. It can be either 'epoch_mean' or 'ema' (exponential moving average).
        ema_decay : float, default 0.99
            The decay to be applied if average_mode is 'ema'.
        f1_mode : float, default 'macro'
            How to calculate the f1-score. Accepts one of these options {binary, macro, weighted}. If binary, then the f1-score
            is calculated only for the positive pixels. If macro, then the f1-score is the average of positive and negative
            scores. If weighted, then the average is weighted according to the number of positive/negative samples.
        """
        super().__init__(dist_sync_on_step=dist_sync_on_step)

        assert average_mode in ['epoch_mean', 'ema']

        self.average_mode = average_mode
        self.prefix = prefix
        self.ema_decay = ema_decay
        self.f1_mode = f1_mode
        self.ema_max_count = min(100, int(1.0/(1.0-ema_decay)))

        self.add_state("epe", default=torch.tensor(0).float(), dist_reduce_fx="sum")
        self.add_state("epe_non_occ", default=torch.tensor(0).float(), dist_reduce_fx="sum")
        self.add_state("epe_occ", default=torch.tensor(0).float(), dist_reduce_fx="sum")

        self.add_state("px1", default=torch.tensor(0).float(), dist_reduce_fx="sum")
        self.add_state("px1_non_occ", default=torch.tensor(0).float(), dist_reduce_fx="sum")
        self.add_state("px1_occ", default=torch.tensor(0).float(), dist_reduce_fx="sum")

        self.add_state("px3", default=torch.tensor(0).float(), dist_reduce_fx="sum")
        self.add_state("px3_non_occ", default=torch.tensor(0).float(), dist_reduce_fx="sum")
        self.add_state("px3_occ", default=torch.tensor(0).float(), dist_reduce_fx="sum")

        self.add_state("px5", default=torch.tensor(0).float(), dist_reduce_fx="sum")
        self.add_state("px5_non_occ", default=torch.tensor(0).float(), dist_reduce_fx="sum")
        self.add_state("px5_occ", default=torch.tensor(0).float(), dist_reduce_fx="sum")

        self.add_state("outlier", default=torch.tensor(0).float(), dist_reduce_fx="sum")
        self.add_state("outlier_non_occ", default=torch.tensor(0).float(), dist_reduce_fx="sum")
        self.add_state("outlier_occ", default=torch.tensor(0).float(), dist_reduce_fx="sum")

        self.add_state("occ_f1", default=torch.tensor(0).float(), dist_reduce_fx="sum")
        self.add_state("mb_f1", default=torch.tensor(0).float(), dist_reduce_fx="sum")
        self.add_state("conf_f1", default=torch.tensor(0).float(), dist_reduce_fx="sum")

        self.add_state("sample_count", default=torch.tensor(0).float(), dist_reduce_fx="sum")
        self.add_state("step_count", default=torch.tensor(0).float(), dist_reduce_fx="sum")

        self.include_occlusion = False

        self.used_keys = []

    def update(
        self,
        preds: Dict[str, torch.Tensor],
        targets: Dict[str, torch.Tensor]
    ) -> None:
        """Compute and update one step of the metrics.

        Parameters
        ----------
        preds : dict[str, torch.Tensor]
            The predictions of the optical flow model.
        targets : dict[str, torch.Tensor]
            The groundtruth of the predictions.
        """
        if self.average_mode == 'epoch_mean':
            prev_weight = 1.0
            next_weight = 1.0
        else:
            prev_weight = self.ema_decay
            next_weight = 1.0 - self.ema_decay

        batch_size = self._get_batch_size(targets['flows'])
        flow_pred = self._fix_shape(preds['flows'], batch_size)
        flow_target = self._fix_shape(targets['flows'], batch_size)

        valid_target = targets.get('valids')
        if valid_target is not None:
            valid_target = self._fix_shape(valid_target, batch_size)
        else:
            valid_target = torch.ones_like(flow_target[:, :1])
        valid_target = valid_target[:, 0]

        occlusion_target = targets.get('occs')
        if occlusion_target is not None:
            occlusion_target = self._fix_shape(occlusion_target, batch_size)

        epe = torch.norm(flow_pred - flow_target, p=2, dim=1)
        px1_mask = (epe < 1).float()
        px3_mask = (epe < 3).float()
        px5_mask = (epe < 5).float()
        target_norm = torch.norm(flow_target, p=2, dim=1)
        outlier_mask = ((epe > 8) & (epe > (0.1*target_norm))).float()
        self.used_keys = [('epe', 'epe', 'valid_target'),
                          ('px1', 'px1_mask', 'valid_target'),
                          ('px3', 'px3_mask', 'valid_target'),
                          ('px5', 'px5_mask', 'valid_target'),
                          ('outlier', 'outlier_mask', 'valid_target')]

        if occlusion_target is not None:
            valid_occ = occlusion_target[:, 0] * valid_target
            valid_non_occ = (1 - occlusion_target[:, 0]) * valid_target
            self.used_keys.extend([('epe_occ', 'epe', 'valid_occ'),
                                   ('epe_non_occ', 'epe', 'valid_non_occ'),
                                   ('px1_occ', 'px1_mask', 'valid_occ'),
                                   ('px1_non_occ', 'px1_mask', 'valid_non_occ'),
                                   ('px3_occ', 'px3_mask', 'valid_occ'),
                                   ('px3_non_occ', 'px3_mask', 'valid_non_occ'),
                                   ('px5_occ', 'px5_mask', 'valid_occ'),
                                   ('px5_non_occ', 'px5_mask', 'valid_non_occ'),
                                   ('outlier_occ', 'outlier_mask', 'valid_occ'),
                                   ('outlier_non_occ', 'outlier_mask', 'valid_non_occ')])
            self.include_occlusion = True

            if preds.get('occs') is not None:
                occlusion_pred = self._fix_shape(preds['occs'], batch_size)
                occ_f1 = self._f1_score(occlusion_pred, occlusion_target, mode=self.f1_mode)
                self.used_keys.extend([('occ_f1', 'occ_f1', 'valid_target')])

        if preds.get('mbs') is not None and targets.get('mbs') is not None:
            mb_pred = self._fix_shape(preds['mbs'], batch_size)
            mb_target = self._fix_shape(targets['mbs'], batch_size)
            mb_f1 = self._f1_score(mb_pred, mb_target, mode=self.f1_mode)
            self.used_keys.extend([('mb_f1', 'mb_f1', 'valid_target')])

        if preds.get('confs') is not None:
            conf_target = torch.exp(-torch.pow(flow_target - flow_pred, 2).sum(dim=1, keepdim=True))
            conf_pred = self._fix_shape(preds['confs'], batch_size)
            conf_f1 = self._f1_score(conf_pred, conf_target, mode=self.f1_mode)
            self.used_keys.extend([('conf_f1', 'conf_f1', 'valid_target')])

        for v1, v2, v3 in self.used_keys:
            setattr(self, v1, prev_weight*getattr(self, v1) + next_weight*self._compute_total(locals()[v2], locals()[v3]))

        self.sample_count += batch_size
        self.step_count += 1

    def calculate_metrics(self) -> Dict[str, torch.Tensor]:
        """Compute and return the average of all metrics.

        On Pytorch-Lightning < 1.2, compute() automatically calls reset(). Sometimes this is not desirable, so the metrics
        are calculated here in this other function, which can be called externally.

        Returns
        -------
        Dict[str, torch.Tensor]
            The average of the metrics.
        """
        if self.average_mode == 'epoch_mean':
            divider = self.sample_count
        else:
            divider = 1.0
            if self.step_count < self.ema_max_count:
                divider -= self.ema_decay ** self.step_count

        metrics = {}
        for k in self.used_keys:
            metrics[self.prefix + k[0]] = getattr(self, k[0]) / divider

        return metrics

    def compute(self) -> Dict[str, torch.Tensor]:
        """Compute and return the average of all metrics.

        Called internally by torchmetrics.

        Returns
        -------
        Dict[str, torch.Tensor]
            The average of the metrics.
        """
        return self.calculate_metrics()

    def _compute_total(
        self,
        tensor: torch.Tensor,
        valid_mask: torch.Tensor
    ) -> torch.Tensor:
        tensor = tensor * valid_mask
        tensor = tensor.view(tensor.shape[0], -1)
        valid_sum = valid_mask.view(valid_mask.shape[0], -1).sum(dim=1)
        valid_sum = torch.clamp(valid_sum, 1)
        tensor = tensor.sum(dim=1) / valid_sum
        if self.average_mode == 'epoch_mean':
            tensor = tensor.sum()
        else:
            tensor = tensor.mean()
        return tensor

    def _f1_score(
        self,
        pred: torch.Tensor,
        target: torch.Tensor,
        mode: str = 'macro'
    ) -> torch.Tensor:
        f1_pos = self._single_f1_score(pred, target)

        if mode == 'binary':
            return f1_pos
        else:
            f1_neg = self._single_f1_score(1-pred, 1-target)
            if mode == 'macro':
                return (f1_pos + f1_neg) / 2.0
            else:  # weighted
                target_pos = (target > 0.5).float()
                target_pos = target_pos.view(target_pos.shape[0], target_pos.shape[1], -1)
                n_pos = target_pos.sum(dim=2)[:, :, None, None]
                w_pos = n_pos / target_pos.shape[2]

                target_neg = (target <= 0.5).float()
                target_neg = target_neg.view(target_neg.shape[0], target_neg.shape[1], -1)
                n_neg = target_neg.sum(dim=2)[:, :, None, None]
                w_neg = n_neg / target_neg.shape[2]

                f1_weighted = w_pos*f1_pos + w_neg*f1_neg

                return f1_weighted

    def _single_f1_score(
        self,
        pred: torch.Tensor,
        target: torch.Tensor
    ) -> torch.Tensor:
        pred_bin = (pred > 0.5).float()
        target_bin = (target > 0.5).float()

        dims = pred_bin.shape
        pred_bin = pred_bin.view(*dims[:-2], -1)
        target_bin = target_bin.view(*dims[:-2], -1)

        tp = (pred_bin * target_bin).sum(dim=-1)
        fp = ((1-pred_bin) * target_bin).sum(dim=-1)
        fn = (pred_bin * (1-target_bin)).sum(dim=-1)

        eps = torch.finfo(pred.dtype).eps
        precision = tp / (tp + fp + eps)
        recall = tp / (tp + fn + eps)

        f1 = 2*precision*recall / (precision + recall + eps)

        return f1[:, :, None]

    def _fix_shape(
        self,
        tensor: torch.Tensor,
        batch_size: int
    ) -> torch.Tensor:
        if len(tensor.shape) == 2:
            tensor = tensor[None, None]
        elif len(tensor.shape) == 3:
            if tensor.shape[0] == batch_size:
                tensor = tensor[:, None]
            else:
                tensor = tensor[None]
        elif len(tensor.shape) == 5:
            tensor = tensor.view(tensor.shape[0]*tensor.shape[1], tensor.shape[2], tensor.shape[3], tensor.shape[4])
        return tensor

    def _get_batch_size(
        self,
        flow_tensor: torch.Tensor
    ) -> int:
        if len(flow_tensor.shape) < 4:
            return 1
        else:
            return flow_tensor.shape[0]



"""Simple utilities to measure elapsed time."""

# =============================================================================
# Copyright 2021 Henrique Morimitsu
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# =============================================================================

import logging
import time
from typing import Union, Tuple

try:
    import torch
except NameError:
    torch = None


class Timer(object):
    """Utility to count the total elapsed time.

    Every time toc() is called, the elapsed time since the last tic() is added to the total time. Call reset() to zero the
    total time.

    Attributes
    ----------
    name : str
        A string name that will be printed to identify this timer.
    indent_level : int, default 0
        The level of indentation when printing this timer. Useful for creating better prints. E.g., inner parts may have a
        larger indentation.

    Examples
    --------
    >>> t1 = Timer('parent_op', 0)
    >>> t1.tic()
    >>> ...
    >>> t2 = Timer('inner_op', 1)
    >>> t2.tic()
    >>> ...
    >>> t2.toc()
    >>> t1.toc()
    >>> print(t1)
    parent_op: 2000.0 (2000.0) ms
    >>> print(t2)
      inner_op: 1000.0 (1000.0) ms
    """

    def __init__(
        self,
        name: str,
        indent_level: int = 0
    ) -> None:
        """Initialize the Timer.

        Parameters
        ----------
        name : str
            A string name that will be printed to identify this timer.
        indent_level : int, optional
            The level of indentation when printing this timer. Useful for creating better prints. E.g., inner parts may have a
            larger indentation.
        """
        self.name = name
        self.indent_level = indent_level
        self.reset()

        self.num_tocs = 0

    def reset(self) -> None:
        """Zero the total time counter."""
        self.total_time = 0.0

    def tic(self) -> None:
        """Start to count the elapsed time."""
        self.has_tic = True
        if torch is not None and torch.cuda.is_available():
            torch.cuda.synchronize()
        self.start = time.perf_counter()

    def toc(self) -> None:
        """Count the elapsed time since the last tic() and add it to the total time."""
        if torch is not None and torch.cuda.is_available():
            torch.cuda.synchronize()
        self.end = time.perf_counter()
        assert self.has_tic, 'toc called without tic'
        self.total_time += self.end - self.start
        self.has_tic = False
        self.num_tocs += 1

    def mean(self) -> float:
        """Return the average time (total time divided by the number of tocs).

        Returns
        -------
        float
            The average time in milliseconds.
        """
        return self.total() / max(1, self.num_tocs)

    def total(self) -> float:
        """Return the total time since the last reset().

        Returns
        -------
        float
            The total time in milliseconds.
        """
        return self.total_time

    def __repr__(self) -> str:
        return f'{"  "*self.indent_level}{self.name}: {1000 * self.total():.1f} ({1000 * self.mean():.1f}) ms'

    def __str__(self) -> str:
        return self.__repr__()


class TimerManager(object):
    """Utility to handle multiple timers.

    Timers can be accessed using a dict-like call (see Usage below). The timers can be either printed to the default output
    or to a log file.

    Attributes
    ----------
    timers : dict[Union[str, tuple[str, int]], Timer]
        The timers to be managed. The dict key can be either a single string representing the name of the timer, or a tuple
        (name, indentation_level)
    log_id : str, default 'timer'
        A string representing the name of this manager.
    log_path : str, default 'timer_log.txt'
        Path to where the log file will be saved (if log is used).
    logger : logging.Logger
        A hander for the logger.

    Examples
    --------
    >>> tm = TimerManager()
    >>> tm['op1'].tic()
    >>> ...
    >>> tm['op1'].toc()
    >>> # You may pass a tuple (str, int) as a key, which will be interpreted
    >>> # as the (name, indent_level) for the timer (see Timer above):
    >>> tm[('op2', 1)].tic()
    >>> ...
    >>> tm['op2'].toc()
    >>> print(tm)  # Prints all timers to default output
    op1: 2000.0 (2000.0) ms
      op2: 1000.0 (1000.0) ms
    >>> tm.write_to_log('Some header message (optional)')  # Write timers to a log file

    See Also
    --------
    Timer : The timers that are managed.
    """

    def __init__(
        self,
        log_id: str = 'timer',
        log_path: str = 'timer_log.txt'
    ) -> None:
        """Initialize the TimerManager.

        Parameters
        ----------
        log_id : str, default 'timer'
            A string representing the name of this manager.
        log_path : str, default 'timer_log.txt'
            Path to where the log file will be saved (if log is used).
        """
        self.timers = {}
        self.log_id = log_id
        self.log_path = log_path
        self.logger = None

    def clear(self) -> None:
        """Remove all timers."""
        self.timers = {}

    def reset(self) -> None:
        """Restart the total time counter of all timers."""
        for _, t in self.timers.items():
            t.reset()

    def write_to_log(
        self,
        header: str = ''
    ) -> None:
        """Write the timers to the log file.

        Parameters
        ----------
        header : str
            An optional string to be added to the top of the log file.
        """
        if self.logger is None:
            self._init_logger()
        if len(header) > 0:
            self.logger.info(header)
        self.logger.info(self.__repr__())

    def _init_logger(self) -> None:
        self.logger = logging.getLogger(self.log_id)
        self.logger.setLevel(logging.INFO)
        fh = logging.FileHandler(self.log_path, mode='w')
        fh.setLevel(logging.INFO)
        self.logger.addHandler(fh)

    def __getitem__(
        self,
        key: Union[str, Tuple[str, int]]
    ) -> None:
        indent_level = 0
        if isinstance(key, tuple) or isinstance(key, list):
            indent_level = key[1]
            key = key[0]
        if self.timers.get(key) is None:
            self.timers[key] = Timer(key, indent_level)
        return self.timers[key]

    def __repr__(self) -> str:
        ret = ''
        for _, t in self.timers.items():
            ret += t.__repr__() + '\n'
        return ret

    def __str__(self) -> str:
        return self.__repr__()



"""

This code is a port to PyTorch of the flow to RGB convertion from flowpy.

https://gitlab-research.centralesupelec.fr/2018seznecm/flowpy

"""

# =============================================================================
# Copyright 2021 Henrique Morimitsu
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# =============================================================================

import math
from collections import namedtuple
from typing import Optional, Tuple, Union

import torch

from .external.flowpy import make_colorwheel

DEFAULT_TRANSITIONS = (15, 6, 4, 11, 13, 6)


def flow_to_rgb(
    flow: torch.Tensor,
    flow_max_radius: Optional[Union[float, torch.Tensor]] = None,
    background: str = 'bright',
    custom_colorwheel: Optional[torch.Tensor] = None
) -> torch.Tensor:
    """Create a RGB representation of an optical flow.

    Parameters
    ----------
    flow : torch.Tensor
        Flow with at least 3 dimensions in the ...CHW (..., Channels, Height, Width) layout, where ... represents any number
        of dimensions.
        flow[..., 0, h, w] should be the x-displacement
        flow[..., 1, h, w] should be the y-displacement
    flow_max_radius : float or torch.Tensor, optional
        Set the radius that gives the maximum color intensity, useful for comparing different flows.
        Default: The normalization is based on the input flow maximum radius per batch element.
    background : str, default 'bright'
        States if zero-valued flow should look 'bright' or 'dark'.
    custom_colorwheel : torch.Tensor
        Use a custom colorwheel for specific hue transition lengths. By default, the default transition lengths are used.

    Returns
    -------
    torch.Tensor
        The RGB representation of the flow. RGB values are float in the [0, 1] interval. The output shape is (..., 3, H, W).

    Raises
    ------
    ValueError
        If the background choice is invalid.

    See Also
    --------
    ptlflow.utils.external.flowpy.make_colorwheel : How the colorwheel can be generated.
    """
    valid_backgrounds = ('bright', 'dark')
    if background not in valid_backgrounds:
        raise ValueError(f'background should be one the following: {valid_backgrounds}, not {background}')

    wheel = make_colorwheel() if custom_colorwheel is None else custom_colorwheel
    wheel = torch.from_numpy(wheel).to(dtype=flow.dtype, device=flow.device) / 255

    orig_shape = flow.shape
    if len(orig_shape) == 3:
        flow = flow[None]
    elif len(orig_shape) > 4:
        flow = flow.view(-1, 2, flow.shape[-2], flow.shape[-1])

    complex_flow = flow[:, 0] + 1j * flow[:, 1]
    complex_flow, nan_mask = _replace_nans(complex_flow)

    radius, angle = torch.abs(complex_flow), torch.angle(complex_flow)

    if flow_max_radius is None:
        flow_max_radius = radius.view(radius.shape[0], -1).max(dim=1)[0]
    else:
        flow_max_radius = torch.zeros(radius.shape[0]).to(dtype=flow.dtype, device=flow.device) + flow_max_radius

    flow_max_radius = torch.clamp(flow_max_radius[:, None, None], 1)
    radius /= flow_max_radius

    ncols = len(wheel)

    # Map the angles from (-pi, pi] to [0, 2pi) to [0, ncols - 1)
    angle[angle < 0] += 2 * math.pi
    angle = angle * ((ncols - 1) / (2 * math.pi))

    # Make the wheel cyclic for interpolation
    wheel = torch.cat((wheel, wheel[:1]), dim=0)

    # Interpolate the hues
    angle_fractional = torch.frac(angle)
    angle_floor = torch.floor(angle)
    angle_ceil = torch.ceil(angle)
    angle_fractional = angle_fractional.view((angle_fractional.shape)+(1,))
    float_hue = (wheel[angle_floor.long()] * (1 - angle_fractional)
                 + wheel[angle_ceil.long()] * angle_fractional)

    ColorizationArgs = namedtuple('ColorizationArgs', [
        'move_hue_valid_radius',
        'move_hue_oversized_radius',
        'invalid_color'])

    def _move_hue_on_v_axis(hues: torch.Tensor, factors: torch.Tensor) -> torch.Tensor:
        return hues * torch.unsqueeze(factors, -1)

    def _move_hue_on_s_axis(hues: torch.Tensor, factors: torch.Tensor) -> torch.Tensor:
        return 1. - torch.unsqueeze(factors, -1) * (1. - hues)

    if background == 'dark':
        parameters = ColorizationArgs(_move_hue_on_v_axis, _move_hue_on_s_axis,
                                      torch.zeros(3).to(dtype=flow.dtype, device=flow.device)+1)
    else:
        parameters = ColorizationArgs(_move_hue_on_s_axis, _move_hue_on_v_axis,
                                      torch.zeros(3).to(dtype=flow.dtype, device=flow.device))

    colors = parameters.move_hue_valid_radius(float_hue, radius)

    oversized_radius_mask = radius > 1
    colors[oversized_radius_mask] = parameters.move_hue_oversized_radius(
        float_hue[oversized_radius_mask],
        1 / radius[oversized_radius_mask]
    )
    colors[nan_mask] = parameters.invalid_color

    output_shape = tuple(3 if i == len(orig_shape)-3 else orig_shape[i] for i in range(len(orig_shape)))
    colors = colors.permute(0, 3, 1, 2).contiguous()
    colors = colors.view(output_shape)

    return colors


def _replace_nans(
    array: torch.Tensor,
    value: int = 0
) -> Tuple[torch.Tensor, torch.Tensor]:
    nan_mask = torch.isnan(array)
    array[nan_mask] = value

    return array, nan_mask



import math
import numpy as np
import torch
from torch import nn
from . import dla

BatchNorm = nn.BatchNorm2d


class Identity(nn.Module):

    def __init__(self):
        super(Identity, self).__init__()

    def forward(self, x):
        return x


def fill_up_weights(up):
    w = up.weight.data
    f = math.ceil(w.size(2) / 2)
    c = (2 * f - 1 - f % 2) / (2. * f)
    for i in range(w.size(2)):
        for j in range(w.size(3)):
            w[0, 0, i, j] = \
                (1 - math.fabs(i / f - c)) * (1 - math.fabs(j / f - c))
    for c in range(1, w.size(0)):
        w[c, 0, :, :] = w[0, 0, :, :]


class IDAUp(nn.Module):

    def __init__(self, node_kernel, out_dim, channels, up_factors):
        super(IDAUp, self).__init__()
        self.channels = channels
        self.out_dim = out_dim
        for i, c in enumerate(channels):
            if c == out_dim:
                proj = Identity()
            else:
                proj = nn.Sequential(
                    nn.Conv2d(c, out_dim, kernel_size=1, stride=1, bias=False),
                    BatchNorm(out_dim), nn.ReLU(inplace=True))
            f = int(up_factors[i])
            if f == 1:
                up = Identity()
            else:
                up = nn.ConvTranspose2d(
                    out_dim,
                    out_dim,
                    f * 2,
                    stride=f,
                    padding=f // 2,
                    output_padding=0,
                    groups=out_dim,
                    bias=False)
                fill_up_weights(up)
            setattr(self, 'proj_' + str(i), proj)
            setattr(self, 'up_' + str(i), up)

        for i in range(1, len(channels)):
            node = nn.Sequential(
                nn.Conv2d(
                    out_dim * 2,
                    out_dim,
                    kernel_size=node_kernel,
                    stride=1,
                    padding=node_kernel // 2,
                    bias=False), BatchNorm(out_dim), nn.ReLU(inplace=True))
            setattr(self, 'node_' + str(i), node)

        for m in self.modules():
            classname = m.__class__.__name__
            if isinstance(m, nn.Conv2d):
                m.weight.data.normal_(0, 0.02)
            elif classname.find('BatchNorm') != -1:
                m.weight.data.fill_(1)
                m.bias.data.zero_()

    def forward(self, layers):
        assert len(self.channels) == len(layers), \
            '{} vs {} layers'.format(len(self.channels), len(layers))
        layers = list(layers)
        for i, l in enumerate(layers):
            upsample = getattr(self, 'up_' + str(i))
            project = getattr(self, 'proj_' + str(i))
            layers[i] = upsample(project(l))
        x = layers[0]
        y = []
        for i in range(1, len(layers)):
            node = getattr(self, 'node_' + str(i))
            x = node(torch.cat([x, layers[i]], 1))
            y.append(x)
        return x, y


class DLAUp(nn.Module):

    def __init__(self, channels, scales=(1, 2, 4, 8, 16), in_channels=None):
        super(DLAUp, self).__init__()
        if in_channels is None:
            in_channels = channels
        self.channels = channels
        channels = list(channels)
        scales = np.array(scales, dtype=int)
        for i in range(len(channels) - 1):
            j = -i - 2
            setattr(
                self, 'ida_{}'.format(i),
                IDAUp(3, channels[j], in_channels[j:],
                      scales[j:] // scales[j]))
            scales[j + 1:] = scales[j]
            in_channels[j + 1:] = [channels[j] for _ in channels[j + 1:]]

    def forward(self, layers):
        layers = list(layers)
        assert len(layers) > 1
        ms_feat = [layers[-1]]
        for i in range(len(layers) - 1):
            ida = getattr(self, 'ida_{}'.format(i))
            x, y = ida(layers[-i - 2:])  # y : aggregation nodes
            layers[-i - 1:] = y
            ms_feat.append(x)
        return ms_feat  # x


class DLAUpEncoder(nn.Module):

    def __init__(self, planes):
        super(DLAUpEncoder, self).__init__()
        self.first_level = 1
        self.base = dla.dla34(planes)
        scales = [2**i for i in range(len(planes[self.first_level:]))]
        self.dla_up = DLAUp(planes[self.first_level:], scales=scales)

    def forward(self, x):
        x = self.base(x)
        y = self.dla_up(x[self.first_level:])
        return y[::-1]


def DLAEncoder(planes):
    model = DLAUpEncoder(planes)
    return model



"""Basic functions to handle optical flow."""

# =============================================================================
# Copyright 2021 Henrique Morimitsu
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# =============================================================================

import pathlib
from typing import IO, Optional, Union

import numpy as np
import torch

from .external import flowpy, raft, selflow
from . import flowpy_torch


def flow_to_rgb(
    flow: Union[np.ndarray, torch.Tensor],
    flow_max_radius: Optional[Union[float, torch.Tensor]] = None,
    background: str = 'bright',
    custom_colorwheel: Optional[torch.Tensor] = None
) -> Union[np.ndarray, torch.Tensor]:
    """Convert flows to RGB images.

    The input can be either numpy or torch tensors. This function is just a wrapper for flowpy and flowpy_torch.

    Parameters
    ----------
    flow : np.ndarray or torch.Tensor
        If flow is a numpy array, then it must have 3 dimensions HWC (Height, Width, Channels) - notice it is channels last.
        If it is a torch tensor, then it has at least 3 dimensions in the ...CHW (..., Channels, Height, Width) layout,
        where ... represents any number of dimensions.
        Channel 0 should be the x-displacement.
        Channel 1 should be the y-displacement.
    flow_max_radius : float or torch.Tensor, optional
        Set the radius that gives the maximum color intensity, useful for comparing different flows.
        Default: The normalization is based on the input flow maximum radius per batch element.
    background : str, default 'bright'
        States if zero-valued flow should look 'bright' or 'dark'.
    custom_colorwheel : np.ndarray or torch.Tensor
        Use a custom colorwheel for specific hue transition lengths. By default, the default transition lengths are used.

    Returns
    -------
    np.ndarray or torch.Tensor
        The RGB image representing the flow. It keeps the same dimensions and type as the input.

    See Also
    --------
    ptlflow.utils.external.flowpy.flow_to_rgb
    ptlflow.utils.flowpy_torch.flow_to_rgb
    """
    if isinstance(flow, np.ndarray):
        flow_rgb = flowpy.flow_to_rgb(flow, flow_max_radius, background, custom_colorwheel)
    else:
        flow_rgb = flowpy_torch.flow_to_rgb(flow, flow_max_radius, background, custom_colorwheel)
    return flow_rgb


def flow_read(
    input_file: Union[str, pathlib.Path, IO],
    format: str = None
) -> np.ndarray:
    """Read optical flow from file.

    This is just a wrapper for flowpy (for .flo and .png) or raft (for pfm), added for convenience.

    Parameters
    ----------
    input_file: str, pathlib.Path or IO
        Path of the file to read or file object.
    format: str, optional
        Specify in what format the flow is read, accepted formats: "png", "flo", or "pfm".
        If None, it is guessed on the file extension.

    Returns
    -------
    numpy.ndarray
        3D flow in the HWF (Height, Width, Flow) layout.
        flow[..., 0] is the x-displacement.
        flow[..., 1] is the y-displacement.

    See Also
    --------
    ptlflow.utils.external.flowpy.flow_read
    ptlflow.utils.external.raft.read_pfm
    write_pfm
    """
    if (format is not None and format == 'pfm') or str(input_file).endswith('pfm'):
        return raft.read_pfm(input_file)
    else:
        return flowpy.flow_read(input_file, format)


def flow_write(
    output_file: Union[str, pathlib.Path, IO],
    flow: np.ndarray,
    format: str = None
) -> None:
    """Write optical flow to file.

    This is just a wrapper for flowpy (for .flo and .png) or selflow (for pfm), added for convenience.

    Parameters
    ----------
    output_file: str, pathlib.Path or IO
        Path of the file to write or file object.
    flow: numpy.ndarray
        3D flow in the HWF (Height, Width, Flow) layout.
        flow[..., 0] should be the x-displacement
        flow[..., 1] should be the y-displacement
    format: str, optional
        Specify in what format the flow is written, accepted formats: "png" or "flo"
        If None, it is guessed on the file extension

    See Also
    --------
    ptlflow.utils.external.flowpy.flow_write
    """
    if (format is not None and format == 'pfm') or str(output_file).endswith('pfm'):
        selflow.write_pfm(output_file, flow)
    else:
        flowpy.flow_write(output_file, flow, format)



'''
Portions of this code copyright 2017, Clement Pinard
'''
from argparse import Namespace
from pathlib import Path

import torch
import torch.nn as nn
from torch.nn import init

from .submodules import *
from .flownet_base import FlowNetBase


class FlowNetS(FlowNetBase):
    pretrained_checkpoints = {
        'things': 'https://github.com/hmorimitsu/ptlflow/releases/download/weights1/flownets-things-98cde14d.ckpt'
    }

    def __init__(self,
                 args: Namespace):
        super(FlowNetS, self).__init__(args)

        self.conv1   = conv(self.args.batch_norm,  self.args.input_channels,   64, kernel_size=7, stride=2)
        self.conv2   = conv(self.args.batch_norm,  64,  128, kernel_size=5, stride=2)
        self.conv3   = conv(self.args.batch_norm, 128,  256, kernel_size=5, stride=2)
        self.conv3_1 = conv(self.args.batch_norm, 256,  256)
        self.conv4   = conv(self.args.batch_norm, 256,  512, stride=2)
        self.conv4_1 = conv(self.args.batch_norm, 512,  512)
        self.conv5   = conv(self.args.batch_norm, 512,  512, stride=2)
        self.conv5_1 = conv(self.args.batch_norm, 512,  512)
        self.conv6   = conv(self.args.batch_norm, 512, 1024, stride=2)
        self.conv6_1 = conv(self.args.batch_norm,1024, 1024)

        self.deconv5 = deconv(1024,512)
        self.deconv4 = deconv(1026,256)
        self.deconv3 = deconv(770,128)
        self.deconv2 = deconv(386,64)

        self.predict_flow6 = predict_flow(1024)
        self.predict_flow5 = predict_flow(1026)
        self.predict_flow4 = predict_flow(770)
        self.predict_flow3 = predict_flow(386)
        self.predict_flow2 = predict_flow(194)

        self.upsampled_flow6_to_5 = nn.ConvTranspose2d(2, 2, 4, 2, 1, bias=False)
        self.upsampled_flow5_to_4 = nn.ConvTranspose2d(2, 2, 4, 2, 1, bias=False)
        self.upsampled_flow4_to_3 = nn.ConvTranspose2d(2, 2, 4, 2, 1, bias=False)
        self.upsampled_flow3_to_2 = nn.ConvTranspose2d(2, 2, 4, 2, 1, bias=False)

        for m in self.modules():
            if isinstance(m, nn.Conv2d):
                if m.bias is not None:
                    init.uniform_(m.bias)
                init.xavier_uniform_(m.weight)

            if isinstance(m, nn.ConvTranspose2d):
                if m.bias is not None:
                    init.uniform_(m.bias)
                init.xavier_uniform_(m.weight)
                # init_deconv_bilinear(m.weight)
        self.upsample1 = nn.Upsample(scale_factor=4, mode='bilinear', align_corners=False)

    def forward(self, inputs):
        x = inputs['images']

        x = x.view(x.shape[0], x.shape[1]*x.shape[2], x.shape[3], x.shape[4])
        out_conv1 = self.conv1(x)

        out_conv2 = self.conv2(out_conv1)
        out_conv3 = self.conv3_1(self.conv3(out_conv2))
        out_conv4 = self.conv4_1(self.conv4(out_conv3))
        out_conv5 = self.conv5_1(self.conv5(out_conv4))
        out_conv6 = self.conv6_1(self.conv6(out_conv5))

        flow6       = self.predict_flow6(out_conv6)
        flow6_up    = self.upsampled_flow6_to_5(flow6)
        out_deconv5 = self.deconv5(out_conv6)
        
        concat5 = torch.cat((out_conv5,out_deconv5,flow6_up),1)
        flow5       = self.predict_flow5(concat5)
        flow5_up    = self.upsampled_flow5_to_4(flow5)
        out_deconv4 = self.deconv4(concat5)
        
        concat4 = torch.cat((out_conv4,out_deconv4,flow5_up),1)
        flow4       = self.predict_flow4(concat4)
        flow4_up    = self.upsampled_flow4_to_3(flow4)
        out_deconv3 = self.deconv3(concat4)
        
        concat3 = torch.cat((out_conv3,out_deconv3,flow4_up),1)
        flow3       = self.predict_flow3(concat3)
        flow3_up    = self.upsampled_flow3_to_2(flow3)
        out_deconv2 = self.deconv2(concat3)

        concat2 = torch.cat((out_conv2,out_deconv2,flow3_up),1)
        flow2 = self.predict_flow2(concat2)

        outputs = {}

        if self.training:
            outputs['flow_preds'] = [flow2.float(),flow3.float(),flow4.float(),flow5.float(),flow6.float()]
            outputs['flows'] = self.args.div_flow*self.upsample1(flow2.float())[:, None]
        else:
            outputs['flows'] = self.args.div_flow*self.upsample1(flow2.float())[:, None]

        return outputs



"""Perform spatial correlation to generate a cost volume.

This is meant to be a simplified version of the SpatialCorrelationSampler from
https://github.com/ClementPinard/Pytorch-Correlation-extension.

This version is implemented purely in PyTorch. However, it only supports correlation with 1x1 kernels.
It is also not as efficient as the original SpatialCorrelationSampler.
"""

# =============================================================================
# Copyright 2021 Henrique Morimitsu
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# =============================================================================

from typing import Tuple, Union

import torch
import torch.nn as nn
import torch.nn.functional as F


def iter_spatial_correlation_sample(
    input1: torch.Tensor,
    input2: torch.Tensor,
    kernel_size: Union[int, Tuple[int, int]] = 1,
    patch_size: Union[int, Tuple[int, int]] = 1,
    stride: Union[int, Tuple[int, int]] = 1,
    padding: Union[int, Tuple[int, int]] = 0,
    dilation: Union[int, Tuple[int, int]] = 1,
    dilation_patch: Union[int, Tuple[int, int]] = 1
) -> torch.Tensor:
    """Apply spatial correlation sampling from input1 to input2 using iteration in PyTorch.

    This docstring is taken and adapted from the original package.

    Every parameter except input1 and input2 can be either single int or a pair of int. For more information about
    Spatial Correlation Sampling, see this page. https://lmb.informatik.uni-freiburg.de/Publications/2015/DFIB15/

    Parameters
    ----------
    input1 : torch.Tensor
        The origin feature map.
    input2 : torch.Tensor
        The target feature map.
    kernel_size : Union[int, Tuple[int, int]], default 1
        Total size of your correlation kernel, in pixels
    patch_size : Union[int, Tuple[int, int]], default 1
        Total size of your patch, determining how many different shifts will be applied.
    stride : Union[int, Tuple[int, int]], default 1
        Stride of the spatial sampler, will modify output height and width.
    padding : Union[int, Tuple[int, int]], default 0
        Padding applied to input1 and input2 before applying the correlation sampling, will modify output height and width.
    dilation : Union[int, Tuple[int, int]], default 1
        Similar to dilation in convolution.
    dilation_patch : Union[int, Tuple[int, int]], default 1
        Step for every shift in patch.

    Returns
    -------
    torch.Tensor
        Result of correlation sampling.

    Raises
    ------
    NotImplementedError
        If kernel_size != 1.
    NotImplementedError
        If dilation != 1.
    """
    # Make inputs be tuples
    kernel_size = (kernel_size, kernel_size) if isinstance(kernel_size, int) else kernel_size
    patch_size = (patch_size, patch_size) if isinstance(patch_size, int) else patch_size
    stride = (stride, stride) if isinstance(stride, int) else stride
    padding = (padding, padding) if isinstance(padding, int) else padding
    dilation = (dilation, dilation) if isinstance(dilation, int) else dilation
    dilation_patch = (dilation_patch, dilation_patch) if isinstance(dilation_patch, int) else dilation_patch

    if kernel_size[0] != 1 or kernel_size[1] != 1:
        raise NotImplementedError('Only kernel_size=1 is supported.')
    if dilation[0] != 1 or dilation[1] != 1:
        raise NotImplementedError('Only dilation=1 is supported.')
    if (patch_size[0] % 2) == 0 or (patch_size[1] % 2) == 0:
        raise NotImplementedError('Only odd patch sizes are supperted.')

    if max(padding) > 0:
        input1 = F.pad(input1, (padding[1], padding[1], padding[0], padding[0]))
        input2 = F.pad(input2, (padding[1], padding[1], padding[0], padding[0]))

    max_displacement = (dilation_patch[0] * (patch_size[0] - 1) // 2, dilation_patch[1] * (patch_size[1] - 1) // 2)
    input2 = F.pad(input2, (max_displacement[1], max_displacement[1], max_displacement[0], max_displacement[0]))

    b, _, h, w = input1.shape
    input1 = input1[:, :, ::stride[0], ::stride[1]]
    sh, sw = input1.shape[2:4]
    corr = torch.zeros(b, patch_size[0], patch_size[1], sh, sw).to(dtype=input1.dtype, device=input1.device)

    for i in range(0, 2*max_displacement[0]+1, dilation_patch[0]):
        for j in range(0, 2*max_displacement[1]+1, dilation_patch[1]):
            p2 = input2[:, :, i:i+h, j:j+w]
            p2 = p2[:, :, ::stride[0], ::stride[1]]
            corr[:, i//dilation_patch[0], j//dilation_patch[1]] = (input1 * p2).sum(dim=1)

    return corr


class IterSpatialCorrelationSampler(nn.Module):
    """Apply spatial correlation sampling from two inputs using iteration in PyTorch."""

    def __init__(
        self,
        kernel_size: Union[int, Tuple[int, int]] = 1,
        patch_size: Union[int, Tuple[int, int]] = 1,
        stride: Union[int, Tuple[int, int]] = 1,
        padding: Union[int, Tuple[int, int]] = 0,
        dilation: Union[int, Tuple[int, int]] = 1,
        dilation_patch: Union[int, Tuple[int, int]] = 1
    ) -> None:
        """Initialize IterSpatialCorrelationSampler.

        Parameters
        ----------
        kernel_size : Union[int, Tuple[int, int]], default 1
            Total size of your correlation kernel, in pixels
        patch_size : Union[int, Tuple[int, int]], default 1
            Total size of your patch, determining how many different shifts will be applied.
        stride : Union[int, Tuple[int, int]], default 1
            Stride of the spatial sampler, will modify output height and width.
        padding : Union[int, Tuple[int, int]], default 0
            Padding applied to input1 and input2 before applying the correlation sampling, will modify output height and width.
        dilation : Union[int, Tuple[int, int]], default 1
            Similar to dilation in convolution.
        dilation_patch : Union[int, Tuple[int, int]], default 1
            Step for every shift in patch.
        """
        super(IterSpatialCorrelationSampler, self).__init__()
        self.kernel_size = kernel_size
        self.patch_size = patch_size
        self.stride = stride
        self.padding = padding
        self.dilation = dilation
        self.dilation_patch = dilation_patch

    def forward(
        self,
        input1: torch.Tensor,
        input2: torch.Tensor
    ) -> torch.Tensor:
        """Compute the correlation sampling from input1 to input2.

        Parameters
        ----------
        input1 : torch.Tensor
            The origin feature map.
        input2 : torch.Tensor
            The target feature map.

        Returns
        -------
        torch.Tensor
            Result of correlation sampling.
        """
        return iter_spatial_correlation_sample(
            input1=input1, input2=input2, kernel_size=self.kernel_size, patch_size=self.patch_size, stride=self.stride,
            padding=self.padding, dilation=self.dilation, dilation_patch=self.dilation_patch)



import torch.nn as nn

BatchNorm = nn.BatchNorm2d


class BasicBlock(nn.Module):

    def __init__(self, in_planes, out_planes):
        super(BasicBlock, self).__init__()
        self.conv1 = nn.Conv2d(
            in_planes,
            out_planes,
            kernel_size=3,
            stride=2,
            padding=1,
            dilation=1,
            bias=False)
        self.bn1 = BatchNorm(out_planes)
        self.conv2 = nn.Conv2d(
            out_planes,
            out_planes,
            kernel_size=3,
            stride=1,
            padding=1,
            dilation=1,
            bias=False)
        self.bn2 = BatchNorm(out_planes)
        self.conv3 = nn.Conv2d(
            out_planes,
            out_planes,
            kernel_size=3,
            stride=1,
            padding=1,
            dilation=1,
            bias=False)
        self.bn3 = BatchNorm(out_planes)
        self.relu = nn.ReLU(inplace=True)

    def forward(self, x):
        out = self.conv1(x)
        out = self.bn1(out)
        out = self.relu(out)

        out = self.conv2(out)
        out = self.bn2(out)
        out = self.relu(out)

        out = self.conv3(out)
        out = self.bn3(out)
        out = self.relu(out)

        return out


class VGG(nn.Module):

    def __init__(self, block, planes):
        super(VGG, self).__init__()
        self.levels = len(planes)
        channels = [3] + planes
        for i in range(self.levels):
            setattr(self, 'block_{}'.format(i),
                    block(channels[i], channels[i + 1]))

        for m in self.modules():
            classname = m.__class__.__name__
            if isinstance(m, nn.Conv2d):
                m.weight.data.normal_(0, 0.02)
            elif classname.find('BatchNorm') != -1:
                m.weight.data.fill_(1)
                m.bias.data.zero_()

    def forward(self, x):
        out = []
        for i in range(self.levels):
            x = getattr(self, 'block_' + str(i))(x)
            out.append(x)
        return out


def VGGEncoder(planes):
    return VGG(BasicBlock, planes)



from __future__ import print_function
import torch
import torch.nn as nn
import torch.utils.data
from torch.autograd import Variable
import torch.nn.functional as F
import math
import numpy as np
import pdb

class residualBlock(nn.Module):
    expansion = 1

    def __init__(self, in_channels, n_filters, stride=1, downsample=None,dilation=1,with_bn=True):
        super(residualBlock, self).__init__()
        if dilation > 1:
            padding = dilation
        else:
            padding = 1

        if with_bn:
            self.convbnrelu1 = conv2DBatchNormRelu(in_channels, n_filters, 3,  stride, padding, dilation=dilation)
            self.convbn2 = conv2DBatchNorm(n_filters, n_filters, 3, 1, 1)
        else:
            self.convbnrelu1 = conv2DBatchNormRelu(in_channels, n_filters, 3,  stride, padding, dilation=dilation,with_bn=False)
            self.convbn2 = conv2DBatchNorm(n_filters, n_filters, 3, 1, 1, with_bn=False)
        self.downsample = downsample
        self.relu = nn.LeakyReLU(0.1, inplace=True)

    def forward(self, x):
        residual = x

        out = self.convbnrelu1(x)
        out = self.convbn2(out)

        if self.downsample is not None:
            residual = self.downsample(x)

        out += residual
        return self.relu(out)

def conv(in_planes, out_planes, kernel_size=3, stride=1, padding=1, dilation=1):   
    return nn.Sequential(
            nn.Conv2d(in_planes, out_planes, kernel_size=kernel_size, stride=stride, 
                        padding=padding, dilation=dilation, bias=True),
            nn.BatchNorm2d(out_planes),
            nn.LeakyReLU(0.1,inplace=True))


class conv2DBatchNorm(nn.Module):
    def __init__(self, in_channels, n_filters, k_size,  stride, padding, dilation=1, with_bn=True):
        super(conv2DBatchNorm, self).__init__()
        bias = not with_bn

        if dilation > 1:
            conv_mod = nn.Conv2d(int(in_channels), int(n_filters), kernel_size=k_size,
                                 padding=padding, stride=stride, bias=bias, dilation=dilation)

        else:
            conv_mod = nn.Conv2d(int(in_channels), int(n_filters), kernel_size=k_size,
                                 padding=padding, stride=stride, bias=bias, dilation=1)


        if with_bn:
            self.cb_unit = nn.Sequential(conv_mod,
                                         nn.BatchNorm2d(int(n_filters)),)
        else:
            self.cb_unit = nn.Sequential(conv_mod,)

    def forward(self, inputs):
        outputs = self.cb_unit(inputs)
        return outputs

class conv2DBatchNormRelu(nn.Module):
    def __init__(self, in_channels, n_filters, k_size,  stride, padding, dilation=1, with_bn=True):
        super(conv2DBatchNormRelu, self).__init__()
        bias = not with_bn
        if dilation > 1:
            conv_mod = nn.Conv2d(int(in_channels), int(n_filters), kernel_size=k_size, 
                                 padding=padding, stride=stride, bias=bias, dilation=dilation)

        else:
            conv_mod = nn.Conv2d(int(in_channels), int(n_filters), kernel_size=k_size, 
                                 padding=padding, stride=stride, bias=bias, dilation=1)

        if with_bn:
            self.cbr_unit = nn.Sequential(conv_mod,
                                          nn.BatchNorm2d(int(n_filters)),
                                          nn.LeakyReLU(0.1, inplace=True),)
        else:
            self.cbr_unit = nn.Sequential(conv_mod,
                                          nn.LeakyReLU(0.1, inplace=True),)

    def forward(self, inputs):
        outputs = self.cbr_unit(inputs)
        return outputs

class pyramidPooling(nn.Module):

    def __init__(self, in_channels, with_bn=True, levels=4):
        super(pyramidPooling, self).__init__()
        self.levels = levels

        self.paths = []
        for i in range(levels):
            self.paths.append(conv2DBatchNormRelu(in_channels, in_channels, 1, 1, 0, with_bn=with_bn))
        self.path_module_list = nn.ModuleList(self.paths)
        self.relu = nn.LeakyReLU(0.1, inplace=True)
    
    def forward(self, x):
        h, w = x.shape[2:]

        k_sizes = []
        strides = []
        for pool_size in np.linspace(1,min(h,w)//2,self.levels,dtype=int):
            k_sizes.append((int(h/pool_size), int(w/pool_size)))
            strides.append((int(h/pool_size), int(w/pool_size)))
        k_sizes = k_sizes[::-1]
        strides = strides[::-1]

        pp_sum = x

        for i, module in enumerate(self.path_module_list):
            out = F.avg_pool2d(x, k_sizes[i], stride=strides[i], padding=0)
            out = module(out)
            out = F.interpolate(out, size=(h,w), mode='bilinear', align_corners=False)
            pp_sum = pp_sum + 1./self.levels*out
        pp_sum = self.relu(pp_sum/2.)

        return pp_sum

class pspnet(nn.Module):
    """
    Modified PSPNet.  https://github.com/meetshah1995/pytorch-semseg/blob/master/ptsemseg/models/pspnet.py
    """
    def __init__(self, is_proj=True,groups=1):
        super(pspnet, self).__init__()
        self.inplanes = 32
        self.is_proj = is_proj

        # Encoder
        self.convbnrelu1_1 = conv2DBatchNormRelu(in_channels=3, k_size=3, n_filters=16,
                                                 padding=1, stride=2)
        self.convbnrelu1_2 = conv2DBatchNormRelu(in_channels=16, k_size=3, n_filters=16,
                                                 padding=1, stride=1)
        self.convbnrelu1_3 = conv2DBatchNormRelu(in_channels=16, k_size=3, n_filters=32,
                                                 padding=1, stride=1)
        # Vanilla Residual Blocks
        self.res_block3 = self._make_layer(residualBlock,64,1,stride=2)
        self.res_block5 = self._make_layer(residualBlock,128,1,stride=2)
        self.res_block6 = self._make_layer(residualBlock,128,1,stride=2)
        self.res_block7 = self._make_layer(residualBlock,128,1,stride=2)
        self.pyramid_pooling = pyramidPooling(128, levels=3)

        # Iconvs
        self.upconv6 = nn.Sequential(nn.Upsample(scale_factor=2, align_corners=False),
                                     conv2DBatchNormRelu(in_channels=128, k_size=3, n_filters=64,
                                                 padding=1, stride=1))
        self.iconv5 = conv2DBatchNormRelu(in_channels=192, k_size=3, n_filters=128,
                                                 padding=1, stride=1)
        self.upconv5 = nn.Sequential(nn.Upsample(scale_factor=2, align_corners=False),
                                     conv2DBatchNormRelu(in_channels=128, k_size=3, n_filters=64,
                                                 padding=1, stride=1))
        self.iconv4 = conv2DBatchNormRelu(in_channels=192, k_size=3, n_filters=128,
                                                 padding=1, stride=1)
        self.upconv4 = nn.Sequential(nn.Upsample(scale_factor=2, align_corners=False),
                                     conv2DBatchNormRelu(in_channels=128, k_size=3, n_filters=64,
                                                 padding=1, stride=1))
        self.iconv3 = conv2DBatchNormRelu(in_channels=128, k_size=3, n_filters=64,
                                                 padding=1, stride=1)
        self.upconv3 = nn.Sequential(nn.Upsample(scale_factor=2, align_corners=False),
                                     conv2DBatchNormRelu(in_channels=64, k_size=3, n_filters=32,
                                                 padding=1, stride=1))
        self.iconv2 = conv2DBatchNormRelu(in_channels=64, k_size=3, n_filters=64,
                                                 padding=1, stride=1)

        if self.is_proj:
            self.proj6 = conv2DBatchNormRelu(in_channels=128,k_size=1,n_filters=128//groups, padding=0,stride=1)
            self.proj5 = conv2DBatchNormRelu(in_channels=128,k_size=1,n_filters=128//groups, padding=0,stride=1)
            self.proj4 = conv2DBatchNormRelu(in_channels=128,k_size=1,n_filters=128//groups, padding=0,stride=1)
            self.proj3 = conv2DBatchNormRelu(in_channels=64, k_size=1,n_filters=64//groups, padding=0,stride=1)
            self.proj2 = conv2DBatchNormRelu(in_channels=64, k_size=1,n_filters=64//groups, padding=0,stride=1)

        for m in self.modules():
            if isinstance(m, nn.Conv2d):
                n = m.kernel_size[0] * m.kernel_size[1] * m.out_channels
                m.weight.data.normal_(0, math.sqrt(2. / n))
                if hasattr(m.bias,'data'):
                    m.bias.data.zero_()
       

    def _make_layer(self, block, planes, blocks, stride=1):
        downsample = None
        if stride != 1 or self.inplanes != planes * block.expansion:
            downsample = nn.Sequential(nn.Conv2d(self.inplanes, planes * block.expansion,
                                                 kernel_size=1, stride=stride, bias=False),
                                       nn.BatchNorm2d(planes * block.expansion),)
        layers = []
        layers.append(block(self.inplanes, planes, stride, downsample))
        self.inplanes = planes * block.expansion
        for i in range(1, blocks):
            layers.append(block(self.inplanes, planes))
        return nn.Sequential(*layers)

    def forward(self, x):
        # H, W -> H/2, W/2
        conv1 = self.convbnrelu1_1(x)
        conv1 = self.convbnrelu1_2(conv1)
        conv1 = self.convbnrelu1_3(conv1)

        ## H/2, W/2 -> H/4, W/4
        pool1 = F.max_pool2d(conv1, 3, 2, 1)

        # H/4, W/4 -> H/16, W/16
        rconv3 = self.res_block3(pool1)
        conv4 = self.res_block5(rconv3)
        conv5 = self.res_block6(conv4)
        conv6 = self.res_block7(conv5)
        conv6 = self.pyramid_pooling(conv6)

        conv6x = F.interpolate(conv6, [conv5.size()[2],conv5.size()[3]],mode='bilinear', align_corners=False)
        concat5 = torch.cat((conv5,self.upconv6[1](conv6x)),dim=1)
        conv5 = self.iconv5(concat5) 

        conv5x = F.interpolate(conv5, [conv4.size()[2],conv4.size()[3]],mode='bilinear', align_corners=False)
        concat4 = torch.cat((conv4,self.upconv5[1](conv5x)),dim=1)
        conv4 = self.iconv4(concat4) 

        conv4x = F.interpolate(conv4, [rconv3.size()[2],rconv3.size()[3]],mode='bilinear', align_corners=False)
        concat3 = torch.cat((rconv3,self.upconv4[1](conv4x)),dim=1)
        conv3 = self.iconv3(concat3) 

        conv3x = F.interpolate(conv3, [pool1.size()[2],pool1.size()[3]],mode='bilinear', align_corners=False)
        concat2 = torch.cat((pool1,self.upconv3[1](conv3x)),dim=1)
        conv2 = self.iconv2(concat2) 

        if self.is_proj:
            proj6 = self.proj6(conv6)
            proj5 = self.proj5(conv5)
            proj4 = self.proj4(conv4)
            proj3 = self.proj3(conv3)
            proj2 = self.proj2(conv2)
            return proj6,proj5,proj4,proj3,proj2
        else:
            return conv6, conv5, conv4, conv3, conv2


class pspnet_s(nn.Module):
    """
    Modified PSPNet.  https://github.com/meetshah1995/pytorch-semseg/blob/master/ptsemseg/models/pspnet.py
    """
    def __init__(self, is_proj=True,groups=1):
        super(pspnet_s, self).__init__()
        self.inplanes = 32
        self.is_proj = is_proj

        # Encoder
        self.convbnrelu1_1 = conv2DBatchNormRelu(in_channels=3, k_size=3, n_filters=16,
                                                 padding=1, stride=2)
        self.convbnrelu1_2 = conv2DBatchNormRelu(in_channels=16, k_size=3, n_filters=16,
                                                 padding=1, stride=1)
        self.convbnrelu1_3 = conv2DBatchNormRelu(in_channels=16, k_size=3, n_filters=32,
                                                 padding=1, stride=1)
        # Vanilla Residual Blocks
        self.res_block3 = self._make_layer(residualBlock,64,1,stride=2)
        self.res_block5 = self._make_layer(residualBlock,128,1,stride=2)
        self.res_block6 = self._make_layer(residualBlock,128,1,stride=2)
        self.res_block7 = self._make_layer(residualBlock,128,1,stride=2)
        self.pyramid_pooling = pyramidPooling(128, levels=3)

        # Iconvs
        self.upconv6 = nn.Sequential(nn.Upsample(scale_factor=2, align_corners=False),
                                     conv2DBatchNormRelu(in_channels=128, k_size=3, n_filters=64,
                                                 padding=1, stride=1))
        self.iconv5 = conv2DBatchNormRelu(in_channels=192, k_size=3, n_filters=128,
                                                 padding=1, stride=1)
        self.upconv5 = nn.Sequential(nn.Upsample(scale_factor=2, align_corners=False),
                                     conv2DBatchNormRelu(in_channels=128, k_size=3, n_filters=64,
                                                 padding=1, stride=1))
        self.iconv4 = conv2DBatchNormRelu(in_channels=192, k_size=3, n_filters=128,
                                                 padding=1, stride=1)
        self.upconv4 = nn.Sequential(nn.Upsample(scale_factor=2, align_corners=False),
                                     conv2DBatchNormRelu(in_channels=128, k_size=3, n_filters=64,
                                                 padding=1, stride=1))
        self.iconv3 = conv2DBatchNormRelu(in_channels=128, k_size=3, n_filters=64,
                                                 padding=1, stride=1)
        #self.upconv3 = nn.Sequential(nn.Upsample(scale_factor=2, align_corners=False),
        #                             conv2DBatchNormRelu(in_channels=64, k_size=3, n_filters=32,
        #                                         padding=1, stride=1))
        #self.iconv2 = conv2DBatchNormRelu(in_channels=64, k_size=3, n_filters=64,
        #                                         padding=1, stride=1)

        if self.is_proj:
            self.proj6 = conv2DBatchNormRelu(in_channels=128,k_size=1,n_filters=128//groups, padding=0,stride=1)
            self.proj5 = conv2DBatchNormRelu(in_channels=128,k_size=1,n_filters=128//groups, padding=0,stride=1)
            self.proj4 = conv2DBatchNormRelu(in_channels=128,k_size=1,n_filters=128//groups, padding=0,stride=1)
            self.proj3 = conv2DBatchNormRelu(in_channels=64, k_size=1,n_filters=64//groups, padding=0,stride=1)
            #self.proj2 = conv2DBatchNormRelu(in_channels=64, k_size=1,n_filters=64//groups, padding=0,stride=1)

        for m in self.modules():
            if isinstance(m, nn.Conv2d):
                n = m.kernel_size[0] * m.kernel_size[1] * m.out_channels
                m.weight.data.normal_(0, math.sqrt(2. / n))
                if hasattr(m.bias,'data'):
                    m.bias.data.zero_()
       

    def _make_layer(self, block, planes, blocks, stride=1):
        downsample = None
        if stride != 1 or self.inplanes != planes * block.expansion:
            downsample = nn.Sequential(nn.Conv2d(self.inplanes, planes * block.expansion,
                                                 kernel_size=1, stride=stride, bias=False),
                                       nn.BatchNorm2d(planes * block.expansion),)
        layers = []
        layers.append(block(self.inplanes, planes, stride, downsample))
        self.inplanes = planes * block.expansion
        for i in range(1, blocks):
            layers.append(block(self.inplanes, planes))
        return nn.Sequential(*layers)

    def forward(self, x):
        # H, W -> H/2, W/2
        conv1 = self.convbnrelu1_1(x)
        conv1 = self.convbnrelu1_2(conv1)
        conv1 = self.convbnrelu1_3(conv1)

        ## H/2, W/2 -> H/4, W/4
        pool1 = F.max_pool2d(conv1, 3, 2, 1)

        # H/4, W/4 -> H/16, W/16
        rconv3 = self.res_block3(pool1)
        conv4 = self.res_block5(rconv3)
        conv5 = self.res_block6(conv4)
        conv6 = self.res_block7(conv5)
        conv6 = self.pyramid_pooling(conv6)

        conv6x = F.interpolate(conv6, [conv5.size()[2],conv5.size()[3]],mode='bilinear', align_corners=False)
        concat5 = torch.cat((conv5,self.upconv6[1](conv6x)),dim=1)
        conv5 = self.iconv5(concat5) 

        conv5x = F.interpolate(conv5, [conv4.size()[2],conv4.size()[3]],mode='bilinear', align_corners=False)
        concat4 = torch.cat((conv4,self.upconv5[1](conv5x)),dim=1)
        conv4 = self.iconv4(concat4) 

        conv4x = F.interpolate(conv4, [rconv3.size()[2],rconv3.size()[3]],mode='bilinear', align_corners=False)
        concat3 = torch.cat((rconv3,self.upconv4[1](conv4x)),dim=1)
        conv3 = self.iconv3(concat3) 

        #conv3x = F.interpolate(conv3, [pool1.size()[2],pool1.size()[3]],mode='bilinear', align_corners=False)
        #concat2 = torch.cat((pool1,self.upconv3[1](conv3x)),dim=1)
        #conv2 = self.iconv2(concat2) 

        if self.is_proj:
            proj6 = self.proj6(conv6)
            proj5 = self.proj5(conv5)
            proj4 = self.proj4(conv4)
            proj3 = self.proj3(conv3)
        #    proj2 = self.proj2(conv2)
        #    return proj6,proj5,proj4,proj3,proj2
            return proj6,proj5,proj4,proj3
        else:
        #    return conv6, conv5, conv4, conv3, conv2
            return conv6, conv5, conv4, conv3


import torch
import torch.nn as nn

BatchNorm = nn.BatchNorm2d


class PreActBlock(nn.Module):
    '''Pre-activation version of the BasicBlock.'''
    expansion = 1

    def __init__(self, in_planes, planes, stride=1, normalize=True):
        super(PreActBlock, self).__init__()
        if normalize:
            self.bn1 = BatchNorm(in_planes)
        self.conv1 = nn.Conv2d(
            in_planes,
            planes,
            kernel_size=3,
            stride=stride,
            padding=1,
            bias=False)
        self.bn2 = BatchNorm(planes)
        self.conv2 = nn.Conv2d(
            planes, planes, kernel_size=3, stride=1, padding=1, bias=False)
        self.relu = nn.ReLU(inplace=True)

        if stride != 1 or in_planes != self.expansion * planes:
            self.shortcut = nn.Sequential(
                nn.Conv2d(
                    in_planes,
                    self.expansion * planes,
                    kernel_size=1,
                    stride=stride,
                    bias=False))

    def forward(self, x):
        out = self.relu(self.bn1(x)) if hasattr(self, 'bn1') else x
        shortcut = self.shortcut(out) if hasattr(self, 'shortcut') else x
        out = self.conv1(out)
        out = self.conv2(self.relu(self.bn2(out)))
        out += shortcut
        return out


class ResBlock(nn.Module):
    expansion = 1

    def __init__(self, in_planes, planes, stride=1):
        super(ResBlock, self).__init__()
        self.conv1 = nn.Conv2d(
            in_planes,
            planes,
            kernel_size=3,
            stride=stride,
            padding=1,
            bias=False)
        self.bn1 = BatchNorm(planes)
        self.conv2 = nn.Conv2d(
            planes, planes, kernel_size=3, stride=1, padding=1, bias=False)
        self.bn2 = BatchNorm(planes)
        self.relu = nn.ReLU(inplace=True)

        self.shortcut = nn.Sequential()
        if stride != 1 or in_planes != self.expansion * planes:
            self.shortcut = nn.Sequential(
                nn.Conv2d(
                    in_planes,
                    self.expansion * planes,
                    kernel_size=1,
                    stride=stride,
                    bias=False), BatchNorm(self.expansion * planes))

    def forward(self, x):
        out = self.relu(self.bn1(self.conv1(x)))
        out = self.bn2(self.conv2(out))
        out += self.shortcut(x)
        out = self.relu(out)
        return out


class ResnetDecoder(nn.Module):

    def __init__(self, inplane, outplane):
        super(ResnetDecoder, self).__init__()
        self.block1 = PreActBlock(inplane, outplane, normalize=False)
        self.block2 = PreActBlock(outplane, outplane, normalize=True)

    def forward(self, x):
        x = self.block1(x)
        out = self.block2(x)
        return out


class HDADecoder(nn.Module):

    def __init__(self, inplane, outplane):
        super(HDADecoder, self).__init__()
        self.block1 = PreActBlock(inplane, outplane, normalize=False)
        self.block2 = PreActBlock(outplane, outplane, normalize=True)
        self.root = nn.Sequential(
            BatchNorm(outplane * 2), nn.ReLU(inplace=True),
            nn.Conv2d(
                outplane * 2, outplane, kernel_size=1, stride=1, bias=False))

    def forward(self, x):
        y1 = self.block1(x)
        y2 = self.block2(y1)
        out = self.root(torch.cat([y1, y2], 1))
        return out
